# Copyright (C) 2020, 2021, Hitachi, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
REST API client class for Hitachi HBSD Driver.

"""

from http import client as httpclient
import socket
import threading

from eventlet import greenthread
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import timeutils
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.connection import HTTPConnection
from requests.packages.urllib3.poolmanager import PoolManager

from cinder import exception
from cinder.i18n import _
from cinder.volume.drivers.hitachi import hbsd_utils as utils
from cinder.volume import volume_utils

_LOCK_TIMEOUT = 2 * 60 * 60
_REST_TIMEOUT = 30
_EXTEND_TIMEOUT = 10 * 60
_EXEC_RETRY_INTERVAL = 5
_DEFAULT_CONNECT_TIMEOUT = 30
_JOB_API_RESPONSE_TIMEOUT = 30 * 60
_GET_API_RESPONSE_TIMEOUT = 30 * 60
_REST_SERVER_BUSY_TIMEOUT = 2 * 60 * 60
_REST_SERVER_RESTART_TIMEOUT = 10 * 60
_REST_SERVER_ERROR_TIMEOUT = 10 * 60
_KEEP_SESSION_LOOP_INTERVAL = 3 * 60
_ANOTHER_LDEV_MAPPED_RETRY_TIMEOUT = 10 * 60

_TCP_KEEPIDLE = 60
_TCP_KEEPINTVL = 15
_TCP_KEEPCNT = 4

_HTTPS = 'https://'

_REST_LOCKED_ERRORS = [
    ('2E11', '2205'),
    ('2E11', '2207'),
]
LDEV_ALREADY_DEFINED = ('2E22', '0001')
NO_AVAILABLE_LDEV_ID = ('2E11', '2209')
INVALID_SNAPSHOT_POOL = ('2E30', '600E')
_MSGID_REST_SERVER_BUSY = ('KART00003-E',)
_MSGID_LOCK_FAILURE = ('KART40050-E', 'KART40051-E', 'KART40052-E')
EXCEED_WWN_MAX = ('B957', '4184')
ANOTHER_LDEV_MAPPED = ('B958', '0947')
REST_NO_RETRY_ERRORS = [
    ('2E10', '9705'),
    ('2E10', '9706'),
    ('2E10', '9707'),
    ('2E11', '8303'),
    ('2E30', '0007'),
    ('B956', '3173'),
    ('B956', '31D7'),
    ('B956', '31D9'),
    ('B957', '4188'),
    ('B958', '015A'),
    ('B958', '015E'),
    LDEV_ALREADY_DEFINED,
    NO_AVAILABLE_LDEV_ID,
    EXCEED_WWN_MAX,
    INVALID_SNAPSHOT_POOL,
]
MSGID_SPECIFIED_OBJECT_DOES_NOT_EXIST = 'KART30013-E'
_REST_NO_RETRY_MESSAGEIDS = [
    MSGID_SPECIFIED_OBJECT_DOES_NOT_EXIST
]

LOG = logging.getLogger(__name__)
MSG = utils.HBSDMsg


def _build_base_url(ip_addr, ip_port):
    return '%(https)s%(ip)s:%(port)s/ConfigurationManager' % {
        'https': _HTTPS,
        'ip': ip_addr,
        'port': ip_port,
    }


class KeepAliveAdapter(HTTPAdapter):

    def __init__(self, conf):
        self.options = HTTPConnection.default_socket_options + [
            (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
            (socket.IPPROTO_TCP, socket.TCP_KEEPIDLE,
             conf.hitachi_rest_tcp_keepidle),
            (socket.IPPROTO_TCP, socket.TCP_KEEPINTVL,
             conf.hitachi_rest_tcp_keepintvl),
            (socket.IPPROTO_TCP, socket.TCP_KEEPCNT,
             conf.hitachi_rest_tcp_keepcnt),
        ]

        super(KeepAliveAdapter, self).__init__()

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = PoolManager(num_pools=connections,
                                       maxsize=maxsize,
                                       block=block,
                                       socket_options=self.options)


class ResponseData(dict):

    def is_json(self):
        return (self['rsp'].content and
                'json' in self['rsp'].headers['Content-Type'])

    def _init_content(self):
        """Set response object."""
        if self.is_json():
            self['rsp_body'] = self['rsp'].json()
        elif self['rsp'].content:
            self['rsp_body'] = self['rsp'].text
        else:
            self['rsp_body'] = None

    def _init_error(self):
        """Set error object"""
        if self['rsp_body'] and 'errorSource' in self['rsp_body']:
            self['errobj'] = self['rsp_body']
        elif self['rsp_body'] and 'error' in self['rsp_body']:
            self['errobj'] = self['rsp_body']['error']
        else:
            self['errobj'] = {}

    def __init__(self, rsp):
        """Initialize instance variables."""
        super(ResponseData, self).__init__()
        self['rsp'] = rsp
        self['status_code'] = rsp.status_code
        self._init_content()
        self._init_error()

    def job_succeeded(self):
        return (self.is_json() and
                self['rsp_body'].get('status') == 'Completed' and
                self['rsp_body'].get('state') == 'Succeeded')

    def get_err_code(self):
        return utils.safe_get_err_code(self['errobj'])

    def get_return_code(self):
        return utils.safe_get_return_code(self['errobj'])

    def is_success(self, ignore_error, ignore_message_id,
                   ignore_return_code, ignore_all_errors=False):
        """Check the success or failure of the response."""
        return (ignore_all_errors or
                self['status_code'] == httpclient.OK or
                (self['status_code'] == httpclient.ACCEPTED and
                 self.job_succeeded()) or
                self.get_err_code() in ignore_error or
                self['errobj'].get('messageId') in ignore_message_id or
                self.get_return_code() in ignore_return_code)

    def is_locked(self):
        """Check if a response is the error of the lock factor."""
        if not self['errobj']:
            return False
        message_id = self['errobj'].get('messageId')
        retcode = self['errobj'].get('errorCode', {}).get('errorCode')
        return (message_id in _MSGID_LOCK_FAILURE or
                self.get_err_code() in _REST_LOCKED_ERRORS or
                retcode == 'EX_EACCES')

    def is_auth_fail(self):
        """Check if a response is an authorization error."""
        return self['status_code'] == httpclient.UNAUTHORIZED

    def get_message_id(self):
        return utils.safe_get_message_id(self['errobj'])

    def is_no_retry_error(self, no_retry_error_code):
        """Check if a response is a no retry error."""
        return (not self.is_auth_fail() and
                ((self['status_code'] not in
                  list(range(200, 300)) + list(range(500, 600))) or
                 self.get_err_code() in no_retry_error_code or
                 self.get_message_id() in _REST_NO_RETRY_MESSAGEIDS))

    def is_rest_server_busy(self):
        """Check if a response is a server busy error."""
        if not self['errobj']:
            return False
        message_id = self['errobj'].get('messageId')
        return (message_id in _MSGID_REST_SERVER_BUSY)

    def get_errobj(self):
        return {
            'errorSource': self['errobj'].get('errorSource', ''),
            'messageId': self['errobj'].get('messageId', ''),
            'message': self['errobj'].get('message', ''),
            'cause': self['errobj'].get('cause', ''),
            'solution': self['errobj'].get('solution', ''),
            'errorCode': self['errobj'].get('errorCode', {}),
        }

    def get_job_result(self):
        return {'job_id': self['rsp_body'].get('jobId', ''),
                'status': self['rsp_body'].get('status', ''),
                'state': self['rsp_body'].get('state', '')}


class RestApiClient():

    def __init__(self, conf, ip_addr, ip_port, storage_device_id,
                 user_id, user_pass, driver_prefix, tcp_keepalive=False,
                 verify=False):
        """Initialize instance variables."""
        self.conf = conf
        self.ip_addr = ip_addr
        self.ip_port = ip_port
        self.storage_id = storage_device_id
        self.storage_info = {}
        self.user_id = user_id
        self.user_pass = user_pass
        self.tcp_keepalive = tcp_keepalive
        self.verify = verify
        self.connect_timeout = self.conf.hitachi_rest_connect_timeout
        self.login_lock = threading.Lock()
        self.keep_session_loop = loopingcall.FixedIntervalLoopingCall(
            self._keep_session)

        self.base_url = _build_base_url(ip_addr, self.ip_port)
        self.object_url = '%(base_url)s/v1/objects/storages/%(storage_id)s' % {
            'base_url': self.base_url,
            'storage_id': self.storage_id,
        }
        self.service_url = '%(base_url)s/v1/%(storage_id)s/services' % {
            'base_url': self.base_url,
            'storage_id': self.storage_id,
        }
        self.headers = {"content-type": "application/json",
                        "accept": "application/json"}
        self.driver_prefix = driver_prefix

    class Session(requests.auth.AuthBase):

        def __init__(self, id, token):
            """Initialize instance variables."""
            self.id = id
            self.token = token

        def __call__(self, req):
            req.headers['Authorization'] = 'Session %(token)s' % {
                'token': self.token,
            }
            return req

    @volume_utils.trace
    def _request(self, method, url, params=None, body=None,
                 async_=False, **kwargs):
        """Transmit the request to REST API server."""
        kwargs.setdefault('ignore_error', [])
        kwargs['no_retry_error'] = (kwargs['ignore_error'] +
                                    REST_NO_RETRY_ERRORS)
        kwargs.setdefault('no_retry', False)
        kwargs.setdefault('do_raise', True)
        kwargs.setdefault('ignore_message_id', [])
        kwargs.setdefault('no_relogin', False)
        kwargs.setdefault('ignore_return_code', [])
        kwargs.setdefault('ignore_all_errors', False)
        kwargs.setdefault('timeout_message', None)
        kwargs.setdefault('no_log', False)
        kwargs.setdefault('timeout', self.conf.hitachi_rest_timeout)

        headers = dict(self.headers)
        if async_:
            read_timeout = self.conf.hitachi_rest_job_api_response_timeout
            headers.update({
                "Response-Max-Wait": str(
                    self.conf.hitachi_rest_job_api_response_timeout),
                "Response-Job-Status": "Completed;"})
        else:
            read_timeout = self.conf.hitachi_rest_get_api_response_timeout

        auth_data = kwargs.get('auth', self.get_my_session())

        timeout = (self.connect_timeout, read_timeout)

        interval = kwargs.get(
            'interval', self.conf.hitachi_exec_retry_interval)
        retry = True
        start_time = timeutils.utcnow()
        watch = timeutils.StopWatch()

        while retry:
            watch.restart()
            try:
                with requests.Session() as session:
                    if self.tcp_keepalive:
                        session.mount(_HTTPS, KeepAliveAdapter(self.conf))
                    rsp = session.request(method, url,
                                          params=params,
                                          json=body,
                                          headers=headers,
                                          auth=auth_data,
                                          timeout=timeout,
                                          verify=self.verify)

            except Exception as e:
                msg = utils.output_log(
                    MSG.REST_SERVER_CONNECT_FAILED,
                    exception=type(e), message=e,
                    method=method, url=url, params=params, body=body)
                message = _(
                    '%(prefix)s error occurred. %(msg)s' % {
                        'prefix': self.driver_prefix,
                        'msg': msg,
                    }
                )
                raise exception.VolumeDriverException(message)

            response = ResponseData(rsp)
            if (response['status_code'] == httpclient.INTERNAL_SERVER_ERROR and
                    kwargs['timeout'] < _REST_SERVER_RESTART_TIMEOUT):
                kwargs['timeout'] = _REST_SERVER_RESTART_TIMEOUT
            if (response['status_code'] == httpclient.SERVICE_UNAVAILABLE and
                    kwargs['timeout'] < _REST_SERVER_ERROR_TIMEOUT):
                kwargs['timeout'] = _REST_SERVER_ERROR_TIMEOUT
            retry, rsp_data, errobj = self._check_rest_api_response(
                response, start_time,
                method=method, url=url, params=params, body=body, **kwargs)
            if retry:
                watch.stop()
                idle = max(interval - watch.elapsed(), 0)
                greenthread.sleep(idle)
                if not kwargs['no_relogin'] and response.is_auth_fail():
                    auth_data = self.get_my_session()

        return rsp_data, errobj

    def _check_rest_api_response(
            self, response, start_time, method=None,
            url=None, params=None, body=None, **kwargs):
        """Check the response from REST API server."""
        rsp_body = response['rsp_body']
        errobj = response['errobj']
        if response.is_locked():
            if (kwargs['no_retry'] or
                    utils.timed_out(
                        start_time, self.conf.hitachi_lock_timeout)):
                msg = utils.output_log(MSG.REST_API_FAILED,
                                       no_log=kwargs['no_log'],
                                       method=method, url=url,
                                       params=params, body=body,
                                       **response.get_errobj())
                if kwargs['do_raise']:
                    message = _(
                        '%(prefix)s error occurred. %(msg)s' % {
                            'prefix': self.driver_prefix,
                            'msg': msg,
                        }
                    )
                    raise exception.VolumeDriverException(
                        message, errobj=errobj)
                return False, rsp_body, errobj
            else:
                LOG.debug("The resource group to which the operation object "
                          "belongs is being locked by other software.")
                return True, rsp_body, errobj

        if response.is_success(kwargs['ignore_error'],
                               kwargs['ignore_message_id'],
                               kwargs['ignore_return_code'],
                               kwargs['ignore_all_errors']):
            return False, rsp_body, errobj

        if (kwargs['no_retry'] and
                response['status_code'] != httpclient.INTERNAL_SERVER_ERROR or
                response.is_no_retry_error(kwargs['no_retry_error'])):
            retry = False
        elif response.is_auth_fail():
            retry = self.relogin(kwargs['no_relogin'])
        else:
            retry = True

        if retry and response.is_rest_server_busy():
            if utils.timed_out(
                    start_time, self.conf.hitachi_rest_server_busy_timeout):
                retry = False
        elif retry and response.get_err_code() in (ANOTHER_LDEV_MAPPED, ):
            if utils.timed_out(
                    start_time,
                    self.conf.hitachi_rest_another_ldev_mapped_retry_timeout):
                LOG.debug(
                    "Another LDEV is already mapped to the specified LUN.")
                retry = False
        elif retry and utils.timed_out(start_time, kwargs['timeout']):
            if kwargs['timeout_message']:
                utils.output_log(kwargs['timeout_message'][0],
                                 **kwargs['timeout_message'][1])
            if response.is_json():
                msg = utils.output_log(MSG.REST_API_TIMEOUT,
                                       no_log=kwargs['no_log'],
                                       method=method, url=url,
                                       params=params, body=body,
                                       **response.get_job_result())
                if errobj:
                    msg = utils.output_log(MSG.REST_API_FAILED,
                                           no_log=kwargs['no_log'],
                                           method=method, url=url,
                                           params=params, body=body,
                                           **response.get_errobj())
            else:
                msg = utils.output_log(MSG.REST_API_HTTP_ERROR,
                                       no_log=kwargs['no_log'],
                                       status_code=response['status_code'],
                                       response_body=rsp_body,
                                       method=method, url=url,
                                       params=params, body=body)
            if kwargs['do_raise']:
                message = _(
                    '%(prefix)s error occurred. %(msg)s' % {
                        'prefix': self.driver_prefix,
                        'msg': msg,
                    }
                )
                raise exception.VolumeDriverException(
                    message, errobj=errobj)
            return False, rsp_body, errobj

        if errobj:
            LOG.debug('ERROR %s', errobj)
        else:
            LOG.debug('ERROR %s', ' '.join(str(rsp_body).splitlines()))

        if not retry:
            if response.is_json():
                msg = utils.output_log(MSG.REST_API_FAILED,
                                       no_log=kwargs['no_log'],
                                       method=method, url=url,
                                       params=params, body=body,
                                       **response.get_errobj())
            else:
                msg = utils.output_log(MSG.REST_API_HTTP_ERROR,
                                       no_log=kwargs['no_log'],
                                       status_code=response['status_code'],
                                       response_body=rsp_body,
                                       method=method, url=url,
                                       params=params, body=body)
            if kwargs['do_raise']:
                message = _(
                    '%(prefix)s error occurred. %(msg)s' % {
                        'prefix': self.driver_prefix,
                        'msg': msg,
                    }
                )
                raise exception.VolumeDriverException(
                    message, errobj=errobj)
        return retry, rsp_body, errobj

    def set_my_session(self, session):
        self.session = session

    def get_my_session(self):
        return getattr(self, 'session', None)

    def _login(self, do_raise=True):
        """Establishes a session and manages the session."""
        url = '%(url)s/sessions' % {
            'url': self.object_url,
        }
        auth = (self.user_id, self.user_pass)
        rsp, err = self._request("POST", url, auth=auth, no_relogin=True,
                                 do_raise=do_raise,
                                 timeout=self.conf.hitachi_lock_timeout)
        if not err:
            self.set_my_session(self.Session(rsp["sessionId"], rsp["token"]))
            return True
        else:
            return False

    def login(self):
        """Establishes a session and manages the session."""
        LOG.debug("Trying to login.")
        return self._login()

    def get_session(self, session_id, **kwargs):
        """Get a session information."""
        url = '%(url)s/sessions/%(id)s' % {
            'url': self.object_url,
            'id': session_id,
        }
        return self._get_object(url, **kwargs)

    def _has_session(self):
        """Check if there is a session managing."""
        has_session = False
        try:
            session = self.get_my_session()
            if session is not None:
                self.get_session(session.id, no_retry=True, no_log=True)
                has_session = True
        except exception.VolumeDriverException as ex:
            LOG.debug('Failed to get session info: %s', ex)
        return has_session

    def relogin(self, no_relogin, no_log=False):
        """Establishes a session again."""
        retry = False
        if not no_relogin:
            with self.login_lock:
                retry = self._has_session()
                if not retry:
                    LOG.debug("Trying to re-login.")
                    retry = self._login(do_raise=False)
                if not retry:
                    utils.output_log(
                        MSG.REST_LOGIN_FAILED,
                        no_log=no_log, user=self.user_id)
        return retry

    def _keep_session(self):
        """Keep a session."""
        LOG.debug('_keep_session thread is started')
        try:
            self.relogin(False, no_log=True)
        except Exception as ex:
            LOG.debug(
                'relogin() in _keep_session() failed. %s', ex)

    def enter_keep_session(self):
        """Begin the keeping of a session."""
        self.keep_session_loop.start(
            self.conf.hitachi_rest_keep_session_loop_interval)
        LOG.debug('enter_keep_session')

    def _get_object(self, url, params=None, **kwargs):
        """Transmit a GET request that appointed object ID."""
        rsp = self._request("GET", url, params=params, **kwargs)[0]
        return rsp if rsp else None

    def _get_objects(self, url, params=None, **kwargs):
        """Transmit a GET request."""
        rsp = self._request("GET", url, params=params, **kwargs)[0]
        return rsp.get("data") if rsp else None

    def _add_object(self, url, body, **kwargs):
        """Transmit a POST request."""
        rsp, errobj = self._request(
            "POST", url, body=body, async_=True, **kwargs)
        if not rsp:
            return None, errobj
        resources = rsp.get('affectedResources')
        if resources:
            return resources[0].split('/')[-1], errobj
        return None, errobj

    def _delete_object(self, url, params=None, body=None, **kwargs):
        """Transmit a DELETE request."""
        self._request("DELETE", url, params=params, body=body, async_=True,
                      **kwargs)

    def _invoke(self, url, body=None, **kwargs):
        """Transmit a PUT request."""
        self._request("PUT", url, body=body, async_=True, **kwargs)

    def get_pools(self, params=None):
        """Get a list of pool information."""
        url = '%(url)s/pools' % {
            'url': self.object_url,
        }
        return self._get_objects(url, params=params)

    def get_pool(self, pool_id, **kwargs):
        """Get a pool information."""
        url = '%(url)s/pools/%(id)s' % {
            'url': self.object_url,
            'id': pool_id,
        }
        return self._get_object(url, **kwargs)

    def get_ldev(self, ldev_id, **kwargs):
        """Get a ldev information."""
        url = '%(url)s/ldevs/%(id)s' % {
            'url': self.object_url,
            'id': ldev_id,
        }
        return self._get_object(url, **kwargs)

    def get_ldevs(self, params=None, **kwargs):
        """Get a list of ldev information."""
        url = '%(url)s/ldevs' % {
            'url': self.object_url,
        }
        return self._get_objects(url, params=params, **kwargs)

    def add_ldev(self, body, **kwargs):
        """Add a ldev information."""
        url = '%(url)s/ldevs' % {
            'url': self.object_url,
        }
        ldev_id = self._add_object(url, body=body, **kwargs)[0]
        return int(ldev_id) if ldev_id else None

    def delete_ldev(self, ldev_id, body=None, **kwargs):
        """Delete a ldev information."""
        url = '%(url)s/ldevs/%(id)s' % {
            'url': self.object_url,
            'id': ldev_id,
        }
        self._delete_object(url, body=body, **kwargs)

    def modify_ldev(self, ldev_id, body):
        """Modify a ldev information."""
        url = '%(url)s/ldevs/%(id)s' % {
            'url': self.object_url,
            'id': ldev_id,
        }
        self._invoke(url, body=body)

    def extend_ldev(self, ldev_id, body):
        """Expand a ldev size."""
        url = '%(url)s/ldevs/%(id)s/actions/%(action)s/invoke' % {
            'url': self.object_url,
            'id': ldev_id,
            'action': 'expand',
        }
        self._invoke(url, body=body, timeout=self.conf.hitachi_extend_timeout)

    def get_ports(self, params=None):
        """Get a list of port information."""
        url = '%(url)s/ports' % {
            'url': self.object_url,
        }
        return self._get_objects(url, params=params)

    def get_port(self, port_id):
        """Get a port information."""
        url = '%(url)s/ports/%(id)s' % {
            'url': self.object_url,
            'id': port_id,
        }
        return self._get_object(url)

    def get_host_grps(self, params=None):
        """Get a list of host group information."""
        url = '%(url)s/host-groups' % {
            'url': self.object_url,
        }
        return self._get_objects(url, params=params)

    def get_host_grp(self, port_id, host_group_number):
        """Get a host group information."""
        url = '%(url)s/host-groups/%(port)s,%(number)d' % {
            'url': self.object_url,
            'port': port_id,
            'number': host_group_number,
        }
        return self._get_object(url)

    def add_host_grp(self, body, **kwargs):
        """Add a host group information."""
        url = '%(url)s/host-groups' % {
            'url': self.object_url,
        }
        host_group_id = self._add_object(url, body=body, **kwargs)[0]
        return int(host_group_id.split(',')[-1]) if host_group_id else None

    def delete_host_grp(self, port_id, host_group_number):
        """Delete a host group information."""
        url = '%(url)s/host-groups/%(port)s,%(number)d' % {
            'url': self.object_url,
            'port': port_id,
            'number': host_group_number,
        }
        self._delete_object(url)

    def modify_host_grp(self, port_id, host_group_number, body, **kwargs):
        """Modify a host group information."""
        url = '%(url)s/host-groups/%(port)s,%(number)d' % {
            'url': self.object_url,
            'port': port_id,
            'number': host_group_number,
        }
        self._invoke(url, body=body, **kwargs)

    def get_hba_wwns(self, port_id, host_group_number):
        """Get a list of wwn information."""
        url = '%(url)s/host-wwns' % {
            'url': self.object_url,
        }
        params = {"portId": port_id, "hostGroupNumber": host_group_number}
        return self._get_objects(url, params=params)

    def get_hba_wwns_by_name(self, port_id, host_group_name):
        """Get a list of wwn information of the specified name."""
        url = '%(url)s/host-wwns' % {
            'url': self.object_url,
        }
        params = {"portId": port_id, "hostGroupName": host_group_name}
        return self._get_objects(url, params=params)

    def add_hba_wwn(self, port_id, host_group_number, host_wwn, **kwargs):
        """Add a wwn information."""
        url = '%(url)s/host-wwns' % {
            'url': self.object_url,
        }
        body = {"hostWwn": host_wwn, "portId": port_id,
                "hostGroupNumber": host_group_number}
        return self._add_object(url, body=body, **kwargs)[0]

    def get_hba_iscsis(self, port_id, host_group_number):
        """Get a list of ISCSI information."""
        url = '%(url)s/host-iscsis' % {
            'url': self.object_url,
        }
        params = {"portId": port_id, "hostGroupNumber": host_group_number}
        return self._get_objects(url, params=params)

    def get_hba_iscsis_by_name(self, port_id, host_group_name):
        """Get a list of ISCSI information of the specified name."""
        url = '%(url)s/host-iscsis' % {
            'url': self.object_url,
        }
        params = {"portId": port_id, "hostGroupName": host_group_name}
        return self._get_objects(url, params=params)

    def add_hba_iscsi(self, port_id, host_group_number, iscsi_name):
        """Add a ISCSI information."""
        url = '%(url)s/host-iscsis' % {
            'url': self.object_url,
        }
        body = {"iscsiName": iscsi_name, "portId": port_id,
                "hostGroupNumber": host_group_number}
        return self._add_object(url, body=body)[0]

    def get_luns(self, port_id, host_group_number,
                 is_basic_lun_information=False):
        """Get a list of lun information."""
        url = '%(url)s/luns' % {
            'url': self.object_url,
        }
        params = {"portId": port_id, "hostGroupNumber": host_group_number,
                  "isBasicLunInformation": is_basic_lun_information}
        return self._get_objects(url, params=params)

    def add_lun(self, port_id, host_group_number, ldev_id, lun=None, **kwargs):
        """Add a lun information."""
        url = '%(url)s/luns' % {
            'url': self.object_url,
        }
        body = {"portId": port_id, "hostGroupNumber": host_group_number,
                "ldevId": ldev_id}
        if lun is not None:
            body['lun'] = lun
        lun_id, errobj = self._add_object(url, body=body, **kwargs)
        return int(lun_id.split(',')[-1]) if lun_id else None, errobj

    def delete_lun(self, port_id, host_group_number, lun, **kwargs):
        """Delete a lun information."""
        url = '%(url)s/luns/%(port)s,%(number)s,%(lun)d' % {
            'url': self.object_url,
            'port': port_id,
            'number': host_group_number,
            'lun': lun,
        }
        self._delete_object(url, **kwargs)

    def get_snapshots(self, params=None):
        """Get a list of snapshot information."""
        url = '%(url)s/snapshots' % {
            'url': self.object_url,
        }
        return self._get_objects(url, params=params)

    def add_snapshot(self, body, **kwargs):
        """Add a snapshot information."""
        url = '%(url)s/snapshots' % {
            'url': self.object_url,
        }
        return self._add_object(url, body=body, **kwargs)[0]

    def delete_snapshot(self, pvol_ldev_id, mu_number, **kwargs):
        """Delete a snapshot information."""
        url = '%(url)s/snapshots/%(pvol)d,%(mu)d' % {
            'url': self.object_url,
            'pvol': pvol_ldev_id,
            'mu': mu_number,
        }
        self._delete_object(url, **kwargs)

    def unassign_snapshot_volume(self, pvol_ldev_id, mu_number, **kwargs):
        """Unassign a snapshot information."""
        url = '%(url)s/snapshots/%(pvol)d,%(mu)d/actions/%(action)s/invoke' % {
            'url': self.object_url,
            'pvol': pvol_ldev_id,
            'mu': mu_number,
            'action': 'unassign-volume',
        }
        self._invoke(url, **kwargs)

    def restore_snapshot(self, pvol_ldev_id, mu_number, body=None):
        """Restore a snapshot information."""
        url = '%(url)s/snapshots/%(pvol)d,%(mu)d/actions/%(action)s/invoke' % {
            'url': self.object_url,
            'pvol': pvol_ldev_id,
            'mu': mu_number,
            'action': 'restore',
        }
        self._invoke(url, body=body)

    def split_snapshotgroup(self, snapshot_group_id):
        url = '%(url)s/snapshot-groups/%(id)s/actions/%(action)s/invoke' % {
            'url': self.object_url,
            'id': snapshot_group_id,
            'action': 'split',
        }
        self._invoke(url)

    def discard_zero_page(self, ldev_id):
        """Return the ldev's no-data pages to the storage pool."""
        url = '%(url)s/ldevs/%(id)s/actions/%(action)s/invoke' % {
            'url': self.object_url,
            'id': ldev_id,
            'action': 'discard-zero-page',
        }
        self._invoke(url)



# Copyright (C) 2020, 2021, Hitachi, Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""REST interface iSCSI module for Hitachi HBSD Driver."""

from oslo_log import log as logging

from cinder.volume.drivers.hitachi import hbsd_rest as rest
from cinder.volume.drivers.hitachi import hbsd_utils as utils

_ISCSI_HMO_REPORT_FULL_PORTAL = 83
_ISCSI_HMO_DISABLE_IO = 91

LOG = logging.getLogger(__name__)
MSG = utils.HBSDMsg


class HBSDRESTISCSI(rest.HBSDREST):
    """REST interface iscsi class for Hitachi HBSD Driver."""

    def _set_target_portal(self, port):
        """Get port info and store it in an instance variable."""
        result = self.client.get_port(port)
        ipv4_addr = result.get('ipv4Address')
        tcp_port = result.get('tcpPort')
        if not ipv4_addr or not tcp_port:
            return False, ipv4_addr, tcp_port
        self.storage_info['portals'][port] = '%(ip)s:%(port)s' % {
            'ip': ipv4_addr,
            'port': tcp_port,
        }
        return True, ipv4_addr, tcp_port

    def connect_storage(self):
        """Prepare for using the storage."""
        target_ports = self.conf.hitachi_target_ports
        compute_target_ports = self.conf.hitachi_compute_target_ports

        super(HBSDRESTISCSI, self).connect_storage()
        # The port type must be ISCSI and the port attributes must contain TAR.
        params = {'portType': 'ISCSI',
                  'portAttributes': 'TAR'}
        port_list = self.client.get_ports(params=params)
        for port in set(target_ports + compute_target_ports):
            if port not in [port_data['portId'] for port_data in port_list]:
                utils.output_log(
                    MSG.INVALID_PORT, port=port, additional_info='(portType, '
                    'portAttributes): not (ISCSI, TAR)')
        for port_data in port_list:
            port = port_data['portId']
            if port not in set(target_ports + compute_target_ports):
                continue
            has_addr = True
            if not port_data['lunSecuritySetting']:
                addr_info = ""
            elif port in set(target_ports + compute_target_ports):
                has_addr, ipv4_addr, tcp_port = self._set_target_portal(port)
                if not has_addr:
                    addr_info = (', ipv4Address: %s, tcpPort: %s' %
                                 (ipv4_addr, tcp_port))
            if not port_data['lunSecuritySetting'] or not has_addr:
                utils.output_log(
                    MSG.INVALID_PORT, port=port,
                    additional_info='portType: %s, lunSecuritySetting: %s%s' %
                    (port_data['portType'], port_data['lunSecuritySetting'],
                     addr_info))
            if not port_data['lunSecuritySetting']:
                continue
            if target_ports and port in target_ports and has_addr:
                self.storage_info['controller_ports'].append(port)
            if (compute_target_ports and port in compute_target_ports and
                    has_addr):
                self.storage_info['compute_ports'].append(port)

        self.check_ports_info()
        utils.output_log(MSG.SET_CONFIG_VALUE,
                         object='port-<IP address:port> list',
                         value=self.storage_info['portals'])

    def create_target_to_storage(self, port, connector, hba_ids):
        """Create an iSCSI target on the specified port."""
        target_name = '%(prefix)s-%(ip)s' % {
            'prefix': self.driver_info['driver_prefix'],
            'ip': connector['ip'],
        }
        body = {'portId': port, 'hostGroupName': target_name}
        if hba_ids:
            body['iscsiName'] = '%(id)s%(suffix)s' % {
                'id': hba_ids,
                'suffix': self.driver_info['target_iqn_suffix'],
            }
        try:
            gid = self.client.add_host_grp(body, no_log=True)
        except Exception:
            params = {'portId': port}
            host_grp_list = self.client.get_host_grps(params)
            for host_grp_data in host_grp_list:
                if host_grp_data['hostGroupName'] == target_name:
                    return target_name, host_grp_data['hostGroupNumber']
            else:
                raise
        return target_name, gid

    def set_hba_ids(self, port, gid, hba_ids):
        """Connect the specified HBA with the specified port."""
        self.client.add_hba_iscsi(port, gid, hba_ids)

    def set_target_mode(self, port, gid):
        """Configure the iSCSI target to meet the environment."""
        body = {'hostMode': 'LINUX/IRIX',
                'hostModeOptions': [_ISCSI_HMO_REPORT_FULL_PORTAL]}
        if self.conf.hitachi_rest_disable_io_wait:
            body['hostModeOptions'].append(_ISCSI_HMO_DISABLE_IO)
        if self.conf.hitachi_host_mode_options:
            for opt in self.conf.hitachi_host_mode_options:
                if int(opt) not in body['hostModeOptions']:
                    body['hostModeOptions'].append(int(opt))
        self.client.modify_host_grp(port, gid, body)

    def _is_host_iqn_registered_in_target(self, port, gid, host_iqn):
        """Check if the specified IQN is registered with iSCSI target."""
        for hba_iscsi in self.client.get_hba_iscsis(port, gid):
            if host_iqn == hba_iscsi['iscsiName']:
                return True
        return False

    def _set_target_info(self, targets, host_grps, iqn):
        """Set the information of the iSCSI target having the specified IQN."""
        for host_grp in host_grps:
            port = host_grp['portId']
            gid = host_grp['hostGroupNumber']
            storage_iqn = host_grp['iscsiName']
            if self._is_host_iqn_registered_in_target(port, gid, iqn):
                targets['info'][port] = True
                targets['list'].append((port, gid))
                targets['iqns'][(port, gid)] = storage_iqn
                return True
        return False

    def _get_host_iqn_registered_in_target_by_name(
            self, port, target_name, host_iqn):
        """Get the information of the iSCSI target having the specified name

        and the specified IQN.
        """
        for hba_iscsi in self.client.get_hba_iscsis_by_name(port, target_name):
            if host_iqn == hba_iscsi['iscsiName']:
                return hba_iscsi
        return None

    def _set_target_info_by_name(self, targets, port, target_name, iqn):
        """Set the information of the iSCSI target having the specified name

        and the specified IQN.
        """
        host_iqn_registered_in_target = (
            self._get_host_iqn_registered_in_target_by_name(
                port, target_name, iqn))
        if host_iqn_registered_in_target:
            gid = host_iqn_registered_in_target['hostGroupNumber']
            storage_iqn = self.client.get_host_grp(port, gid)['iscsiName']
            targets['info'][port] = True
            targets['list'].append((port, gid))
            targets['iqns'][(port, gid)] = storage_iqn
            return True
        return False

    def find_targets_from_storage(self, targets, connector, target_ports):
        """Find mapped ports, memorize them and return unmapped port count."""
        iqn = self.get_hba_ids_from_connector(connector)
        not_found_count = 0
        for port in target_ports:
            targets['info'][port] = False
            if 'ip' in connector:
                target_name = '%(prefix)s-%(ip)s' % {
                    'prefix': self.driver_info['driver_prefix'],
                    'ip': connector['ip'],
                }
                if self._set_target_info_by_name(
                        targets, port, target_name, iqn):
                    continue
            host_grps = self.client.get_host_grps({'portId': port})
            if 'ip' in connector:
                host_grps = [hg for hg in host_grps
                             if hg['hostGroupName'] != target_name]
            if self._set_target_info(targets, host_grps, iqn):
                pass
            else:
                not_found_count += 1
        return not_found_count

    def initialize_connection(self, volume, connector, is_snapshot=False):
        """Initialize connection between the server and the volume."""
        conn_info, map_info = super(HBSDRESTISCSI, self).initialize_connection(
            volume, connector, is_snapshot)
        return conn_info

    def get_properties_iscsi(self, targets, multipath):
        """Return iSCSI-specific server-LDEV connection info."""
        if not multipath:
            target_list = targets['list'][:1]
        else:
            target_list = targets['list'][:]

        for target in target_list:
            if target not in targets['iqns']:
                port, gid = target
                target_info = self.client.get_host_grp(port, gid)
                iqn = target_info.get('iscsiName') if target_info else None
                if not iqn:
                    msg = utils.output_log(MSG.RESOURCE_NOT_FOUND,
                                           resource='Target IQN')
                    self.raise_error(msg)
                targets['iqns'][target] = iqn
                LOG.debug(
                    'Found target iqn of host group. (port: %(port)s, '
                    'gid: %(gid)s, target iqn: %(iqn)s)',
                    {'port': port, 'gid': gid, 'iqn': iqn})
        return super(HBSDRESTISCSI, self).get_properties_iscsi(
            targets, multipath)

    def _get_iqn(self, port, hostgroup):
        """Get IQN from a port and the ISCSI target."""
        hba_iscsis = self.client.get_hba_iscsis_by_name(port, hostgroup)
        return hba_iscsis[0]['iscsiName']

    def set_terminate_target(self, fake_connector, port_hostgroup_map):
        """Set necessary information in connector in terminate."""
        for port, hostgroups in port_hostgroup_map.items():
            for hostgroup in hostgroups:
                iqn = self._get_iqn(port, hostgroup)
                if iqn:
                    fake_connector['initiator'] = iqn
                    return



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt

from cinder.api import microversions as mv
from cinder.api.v3 import groups
from cinder import exception
from cinder.objects import fields
from cinder.policies import group_actions as group_policies
from cinder.tests.unit.api import fakes as fake_api
from cinder.tests.unit.policies import base
from cinder.tests.unit import utils as test_utils
from cinder.volume import group_types


@ddt.ddt
class GroupActionPolicyTest(base.BasePolicyTest):
    sysadmins = [
        'legacy_admin',
        'project_admin',
        'system_admin',
    ]
    non_sysadmins = [
        'legacy_owner',
        'project_member',
        'project_reader',
        'project_foo',
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_users = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]
    unauthorized_users = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_members = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]
    unauthorized_members = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]
    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.

    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)
        self.controller = groups.GroupsController()
        self.api_path = '/v3/%s/groups' % (self.project_id)
        self.api_version = mv.GROUP_REPLICATION
        self.group_type = group_types.create(self.project_admin_context,
                                             'group_type_name',
                                             {'key3': 'value3'},
                                             is_public=True)
        # not surprisingly, to do a group action you need to get a
        # group, so relax the group:get policy so that these tests
        # will check the group action policy we're interested in
        self.policy.set_rules({"group:get": ""},
                              overwrite=False)

    def _create_group(self, group_status=fields.GroupStatus.AVAILABLE):
        volume_type = test_utils.create_volume_type(self.project_admin_context,
                                                    name="test")
        group = test_utils.create_group(self.project_admin_context,
                                        status=group_status,
                                        group_type_id=self.group_type.id,
                                        volume_type_ids=[volume_type.id])

        test_utils.create_volume(self.project_member_context,
                                 group_id=group.id,
                                 testcase_instance=self,
                                 volume_type_id=volume_type.id)
        return group.id

    @ddt.data(*base.all_users)
    @mock.patch('cinder.group.api.API.enable_replication')
    def test_enable_group_replication_policy(self, user_id,
                                             mock_enable_replication):
        """Test enable group replication policy."""

        # FIXME: this is a very fragile approach
        def fake_enable_rep(context, group):
            context.authorize(group_policies.ENABLE_REP, target_obj=group)

        volume_type = test_utils.create_volume_type(self.project_admin_context,
                                                    name='test_group_policy')

        group = test_utils.create_group(self.project_admin_context,
                                        status=fields.GroupStatus.AVAILABLE,
                                        group_type_id=self.group_type.id,
                                        volume_type_ids=[volume_type.id])

        test_utils.create_volume(self.project_member_context,
                                 group_id=group.id,
                                 testcase_instance=self,
                                 volume_type_id=volume_type.id)

        mock_enable_replication.side_effect = fake_enable_rep
        self.group_type.status = 'enabled'
        rule_name = group_policies.ENABLE_REP
        version = mv.GROUP_REPLICATION
        url = '%s/%s/action' % (self.api_path, group.id)
        req = fake_api.HTTPRequest.blank(url, version=version)
        req.method = 'POST'
        body = {
            "enable_replication": {}
        }
        unauthorized_exceptions = [exception.GroupNotFound]
        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller.enable_replication,
                                 req, id=group.id, body=body)
        group.destroy()

    @ddt.data(*base.all_users)
    @mock.patch('cinder.group.api.API.disable_replication')
    def test_disable_group_replication_policy(self, user_id,
                                              mock_disable_replication):
        """Test disable group replication policy."""

        # FIXME: this is a very fragile approach
        def fake_disable_rep(context, group):
            context.authorize(group_policies.DISABLE_REP, target_obj=group)

        volume_type = test_utils.create_volume_type(self.project_admin_context,
                                                    name='test_group_policy')

        group = test_utils.create_group(self.project_admin_context,
                                        status=fields.GroupStatus.AVAILABLE,
                                        group_type_id=self.group_type.id,
                                        volume_type_ids=[volume_type.id])

        test_utils.create_volume(self.project_member_context,
                                 group_id=group.id,
                                 testcase_instance=self,
                                 volume_type_id=volume_type.id)

        mock_disable_replication.side_effect = fake_disable_rep
        rule_name = group_policies.DISABLE_REP
        version = mv.GROUP_REPLICATION
        url = '%s/%s/action' % (self.api_path, group.id)
        req = fake_api.HTTPRequest.blank(url, version=version)
        req.method = 'POST'
        body = {
            "disable_replication": {}
        }
        unauthorized_exceptions = [exception.GroupNotFound]

        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller.disable_replication,
                                 req, id=group.id, body=body)
        group.destroy()

    @ddt.data(*base.all_users)
    def test_reset_status_group_policy(self, user_id):
        """Test reset status of group policy."""
        rule_name = group_policies.RESET_STATUS
        group_id = self._create_group(group_status=fields.GroupStatus.ERROR)
        url = '%s/%s/action' % (self.api_path, group_id)
        version = mv.GROUP_VOLUME_RESET_STATUS
        req = fake_api.HTTPRequest.blank(url, version=version)
        req.method = 'POST'
        body = {
            "reset_status": {
                "status": "available"
            }
        }
        unauthorized_exceptions = [exception.GroupNotFound]
        self.common_policy_check(user_id,
                                 self.sysadmins,
                                 self.non_sysadmins,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller.reset_status,
                                 req,
                                 id=group_id,
                                 body=body)

    @ddt.data(*base.all_users)
    def test_delete_group_policy(self, user_id):
        """Test delete group policy."""
        volume_type = test_utils.create_volume_type(self.project_admin_context,
                                                    name='test_group_policy')

        group_1 = test_utils.create_group(self.project_admin_context,
                                          status=fields.GroupStatus.AVAILABLE,
                                          group_type_id=self.group_type.id,
                                          volume_type_ids=[volume_type.id])

        rule_name = group_policies.DELETE_POLICY
        url = '%s/%s' % (self.api_path, group_1.id)
        req = fake_api.HTTPRequest.blank(url, version=mv.GROUP_VOLUME)
        req.method = 'POST'
        body = {
            "delete": {
                "delete-volumes": "false"
            }
        }
        unauthorized_exceptions = [exception.GroupNotFound]
        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.delete_group,
                                 req, id=group_1.id, body=body)

    @ddt.data(*base.all_users)
    @mock.patch('cinder.group.api.API.failover_replication')
    def test_fail_over_replication_group_policy(self, user_id,
                                                mock_failover_replication):
        """Test fail over replication group policy."""

        # FIXME: this is a very fragile approach
        def fake_failover_rep(context, group,
                              allow_attached_volume=False,
                              secondary_backend_id=None):
            context.authorize(group_policies.FAILOVER_REP, target_obj=group)

        volume_type = test_utils.create_volume_type(self.project_admin_context,
                                                    name='test_group_policy')

        group_2 = test_utils.create_group(self.project_admin_context,
                                          status=fields.GroupStatus.AVAILABLE,
                                          group_type_id=self.group_type.id,
                                          volume_type_ids=[volume_type.id])

        mock_failover_replication.side_effect = fake_failover_rep
        rule_name = group_policies.FAILOVER_REP
        url = '%s/%s' % (self.api_path, group_2.id)
        req = fake_api.HTTPRequest.blank(url, version=mv.GROUP_REPLICATION)
        req.method = 'POST'
        body = {
            "failover_replication": {
                "allow_attached_volume": "true",
                "secondary_backend_id": "vendor-id-1"
            }
        }
        unauthorized_exceptions = [exception.GroupNotFound]
        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller.failover_replication,
                                 req, id=group_2.id, body=body)

    @ddt.data(*base.all_users)
    @mock.patch('cinder.group.api.API.list_replication_targets')
    def test_list_replication_targets_group_policy(self, user_id,
                                                   mock_list_targets):
        """Test list replication targets for a group policy."""

        # FIXME: this is a very fragile approach
        def fake_list_targets(context, group):
            context.authorize(group_policies.LIST_REP, target_obj=group)

        volume_type = test_utils.create_volume_type(self.project_admin_context,
                                                    name='test_group_policy')

        group_2 = test_utils.create_group(self.project_admin_context,
                                          status=fields.GroupStatus.AVAILABLE,
                                          group_type_id=self.group_type.id,
                                          volume_type_ids=[volume_type.id])

        mock_list_targets.side_effect = fake_list_targets
        rule_name = group_policies.LIST_REP
        url = '%s/%s/action' % (self.api_path, group_2.id)
        req = fake_api.HTTPRequest.blank(url, version=mv.GROUP_REPLICATION)
        req.method = 'POST'
        body = {"list_replication_targets": {}}
        unauthorized_exceptions = [exception.GroupNotFound]
        self.common_policy_check(user_id, self.authorized_members,
                                 self.unauthorized_members,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller.list_replication_targets,
                                 req, id=group_2.id, body=body)
        group_2.destroy()


class GroupActionPolicySecureRbacTest(GroupActionPolicyTest):
    sysadmins = [
        'legacy_admin',
        'system_admin',
        'project_admin',
    ]
    non_sysadmins = [
        'legacy_owner',
        'project_member',
        'system_member',
        'system_reader',
        'system_foo',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]
    authorized_users = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
    ]
    unauthorized_users = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_members = authorized_users
    unauthorized_members = unauthorized_users

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The consistencygroups V3 API."""
from http import HTTPStatus

from oslo_log import log as logging
import webob
from webob import exc

from cinder.api.contrib import consistencygroups as cg_v2
from cinder.api import microversions as mv
from cinder.api.openstack import wsgi
from cinder.i18n import _
from cinder.policies import groups as group_policy

LOG = logging.getLogger(__name__)


class ConsistencyGroupsController(cg_v2.ConsistencyGroupsController):
    """The ConsistencyGroups API controller for the OpenStack API V3."""

    def _check_update_parameters_v3(self, req, name, description, add_volumes,
                                    remove_volumes):
        allow_empty = req.api_version_request.matches(
            mv.CG_UPDATE_BLANK_PROPERTIES, None)
        if allow_empty:
            if (name is None and description is None
                    and not add_volumes and not remove_volumes):
                msg = _("Must specify one or more of the following keys to "
                        "update: name, description, "
                        "add_volumes, remove_volumes.")
                raise exc.HTTPBadRequest(explanation=msg)
        else:
            if not (name or description or add_volumes or remove_volumes):
                msg = _("Name, description, add_volumes, and remove_volumes "
                        "can not be all empty in the request body.")
                raise exc.HTTPBadRequest(explanation=msg)
        return allow_empty

    def update(self, req, id, body):
        """Update the consistency group.

        Expected format of the input parameter 'body':

        .. code-block:: json

            {
                "consistencygroup":
                {
                    "name": "my_cg",
                    "description": "My consistency group",
                    "add_volumes": "volume-uuid-1,volume-uuid-2,...",
                    "remove_volumes": "volume-uuid-8,volume-uuid-9,..."
                }
            }

        """
        LOG.debug('Update called for consistency group %s.', id)
        if not body:
            msg = _("Missing request body.")
            raise exc.HTTPBadRequest(explanation=msg)

        self.assert_valid_body(body, 'consistencygroup')
        context = req.environ['cinder.context']
        group = self._get(context, id)
        context.authorize(group_policy.UPDATE_POLICY, target_obj=group)
        consistencygroup = body.get('consistencygroup', None)
        self.validate_name_and_description(consistencygroup)
        name = consistencygroup.get('name', None)
        description = consistencygroup.get('description', None)
        add_volumes = consistencygroup.get('add_volumes', None)
        remove_volumes = consistencygroup.get('remove_volumes', None)

        allow_empty = self._check_update_parameters_v3(req, name,
                                                       description,
                                                       add_volumes,
                                                       remove_volumes)
        self._update(context, group, name, description, add_volumes,
                     remove_volumes, allow_empty)
        return webob.Response(status_int=HTTPStatus.ACCEPTED)


def create_resource():
    return wsgi.Resource(ConsistencyGroupsController())



# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Tests for encryption key transfer."""

from unittest import mock

from castellan.common.credentials import keystone_password
from oslo_config import cfg

from cinder.common import constants
from cinder import context
from cinder.keymgr import conf_key_mgr
from cinder.keymgr import transfer
from cinder import objects
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test
from cinder.tests.unit import utils as test_utils

CONF = cfg.CONF

ENCRYPTION_SECRET = 'the_secret'
CINDER_USERNAME = 'cinder'
CINDER_PASSWORD = 'key_transfer_test'


class KeyTransferTestCase(test.TestCase):
    OLD_ENCRYPTION_KEY_ID = fake.ENCRYPTION_KEY_ID
    NEW_ENCRYPTION_KEY_ID = fake.ENCRYPTION_KEY2_ID

    key_manager_class = ('castellan.key_manager.barbican_key_manager.'
                         'BarbicanKeyManager')

    def setUp(self):
        super(KeyTransferTestCase, self).setUp()
        self.conf = CONF
        self.conf.set_override('backend',
                               self.key_manager_class,
                               group='key_manager')
        self.conf.set_override('username',
                               CINDER_USERNAME,
                               group='keystone_authtoken')
        self.conf.set_override('password',
                               CINDER_PASSWORD,
                               group='keystone_authtoken')

        self.context = context.RequestContext(fake.USER_ID, fake.PROJECT_ID)

    def _create_volume_and_snapshots(self):
        volume = test_utils.create_volume(
            self.context,
            testcase_instance=self,
            encryption_key_id=self.OLD_ENCRYPTION_KEY_ID)

        _ = test_utils.create_snapshot(
            self.context,
            volume.id,
            display_name='snap_1',
            testcase_instance=self,
            encryption_key_id=self.OLD_ENCRYPTION_KEY_ID)

        _ = test_utils.create_snapshot(
            self.context,
            volume.id,
            display_name='snap_2',
            testcase_instance=self,
            encryption_key_id=self.OLD_ENCRYPTION_KEY_ID)

        return volume

    def _verify_service_context(self, mocked_call):
        service_context = mocked_call.call_args.args[0]
        self.assertIsInstance(service_context,
                              keystone_password.KeystonePassword)
        self.assertEqual(service_context.username, CINDER_USERNAME)
        self.assertEqual(service_context.password, CINDER_PASSWORD)

    def _verify_encryption_key_id(self, volume_id, encryption_key_id):
        volume = objects.Volume.get_by_id(self.context, volume_id)
        self.assertEqual(volume.encryption_key_id, encryption_key_id)

        snapshots = objects.snapshot.SnapshotList.get_all_for_volume(
            self.context, volume.id)
        self.assertEqual(len(snapshots), 2)
        for snapshot in snapshots:
            self.assertEqual(snapshot.encryption_key_id, encryption_key_id)

    def _test_transfer_from_user_to_cinder(self, transfer_fn):
        volume = self._create_volume_and_snapshots()
        with mock.patch(
                self.key_manager_class + '.get',
                return_value=ENCRYPTION_SECRET) as mock_key_get, \
            mock.patch(
                self.key_manager_class + '.store',
                return_value=self.NEW_ENCRYPTION_KEY_ID) as mock_key_store, \
            mock.patch(
                self.key_manager_class + '.delete') as mock_key_delete:

            transfer_fn(self.context, volume)

            # Verify the user's context was used to fetch and delete the
            # volume's current key ID.
            mock_key_get.assert_called_once_with(
                self.context, self.OLD_ENCRYPTION_KEY_ID)
            mock_key_delete.assert_called_once_with(
                self.context, self.OLD_ENCRYPTION_KEY_ID)

            # Verify the cinder service created the new key ID.
            mock_key_store.assert_called_once_with(
                mock.ANY, ENCRYPTION_SECRET)
            self._verify_service_context(mock_key_store)

        # Verify the volume (and its snaps) reference the new key ID.
        self._verify_encryption_key_id(volume.id, self.NEW_ENCRYPTION_KEY_ID)

    def _test_transfer_from_cinder_to_user(self, transfer_fn):
        volume = self._create_volume_and_snapshots()
        with mock.patch(
                self.key_manager_class + '.get',
                return_value=ENCRYPTION_SECRET) as mock_key_get, \
            mock.patch(
                self.key_manager_class + '.store',
                return_value=self.NEW_ENCRYPTION_KEY_ID) as mock_key_store, \
            mock.patch(
                self.key_manager_class + '.delete') as mock_key_delete:

            transfer_fn(self.context, volume)

            # Verify the cinder service was used to fetch and delete the
            # volume's current key ID.
            mock_key_get.assert_called_once_with(
                mock.ANY, self.OLD_ENCRYPTION_KEY_ID)
            self._verify_service_context(mock_key_get)

            mock_key_delete.assert_called_once_with(
                mock.ANY, self.OLD_ENCRYPTION_KEY_ID)
            self._verify_service_context(mock_key_delete)

            # Verify the user's context created the new key ID.
            mock_key_store.assert_called_once_with(
                self.context, ENCRYPTION_SECRET)

        # Verify the volume (and its snaps) reference the new key ID.
        self._verify_encryption_key_id(volume.id, self.NEW_ENCRYPTION_KEY_ID)

    def test_transfer_create(self):
        self._test_transfer_from_user_to_cinder(transfer.transfer_create)

    def test_transfer_accept(self):
        self._test_transfer_from_cinder_to_user(transfer.transfer_accept)

    def test_transfer_delete(self):
        self._test_transfer_from_cinder_to_user(transfer.transfer_delete)


class KeyTransferFixedKeyTestCase(KeyTransferTestCase):
    OLD_ENCRYPTION_KEY_ID = constants.FIXED_KEY_ID
    NEW_ENCRYPTION_KEY_ID = constants.FIXED_KEY_ID

    key_manager_class = 'cinder.keymgr.conf_key_mgr.ConfKeyManager'

    def setUp(self):
        super(KeyTransferFixedKeyTestCase, self).setUp()
        self.conf.register_opts(conf_key_mgr.key_mgr_opts, group='key_manager')
        self.conf.set_override('fixed_key',
                               'df393fca58657e6dc76a6fea31c3e7e0',
                               group='key_manager')



#  Copyright (C) 2021-2022 YADRO.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from unittest import mock
from unittest.mock import MagicMock
from unittest.mock import Mock
from unittest import TestCase

from cinder.cmd import manage as cinder_manage
from cinder.exception import ManageExistingInvalidReference
from cinder.exception import VolumeBackendAPIException
from cinder.tests.unit.fake_constants import VOLUME_NAME
from cinder.volume import configuration
from cinder.volume.drivers.yadro.tatlin_client import TatlinAccessAPI
from cinder.volume.drivers.yadro.tatlin_client import TatlinClientCommon
from cinder.volume.drivers.yadro.tatlin_common import tatlin_opts
from cinder.volume.drivers.yadro.tatlin_common import TatlinCommonVolumeDriver
from cinder.volume.drivers.yadro.tatlin_exception import TatlinAPIException
from cinder.volume.drivers.yadro.tatlin_utils import TatlinVolumeConnections


OSMGR_ISCSI_PORTS = [
    {
        "id": "ip-sp-1-98039b04091a",
        "meta": {
            "tatlin-node": "sp-1",
            "type": "ip",
            "port-type": "active"
        },
        "params": {
            "dhcp": False,
            "ifname": "p30",
            "physical-port": "p30",
            "ipaddress": "172.20.101.65",
            "netmask": "24",
            "mtu": "1500",
            "gateway": "172.20.101.1",
            "roles": "",
            "iflabel": "",
            "wwpn": ""
        }
    },
    {
        "id": "ip-sp-0-b8599f1caf1b",
        "meta": {
            "tatlin-node": "sp-0",
            "type": "ip",
            "port-type": "active"
        },
        "params": {
            "dhcp": False,
            "ifname": "p31",
            "physical-port": "p31",
            "ipaddress": "172.20.101.66",
            "netmask": "24",
            "mtu": "1500",
            "gateway": "172.20.101.1",
            "roles": "",
            "iflabel": "",
            "wwpn": ""
        }
    },
    {
        "id": "ip-sp-1-98039b04091b",
        "meta": {
            "tatlin-node": "sp-1",
            "type": "ip",
            "port-type": "active"
        },
        "params": {
            "dhcp": False,
            "ifname": "p31",
            "physical-port": "p31",
            "ipaddress": "172.20.101.67",
            "netmask": "24",
            "mtu": "1500",
            "gateway": "172.20.101.1",
            "roles": "",
            "iflabel": "",
            "wwpn": ""
        }
    },
    {
        "id": "ip-sp-0-b8599f1caf1a",
        "meta": {
            "tatlin-node": "sp-0",
            "type": "ip",
            "port-type": "active"
        },
        "params": {
            "dhcp": False,
            "ifname": "p30",
            "physical-port": "p30",
            "ipaddress": "172.20.101.64",
            "netmask": "24",
            "mtu": "1500",
            "gateway": "172.20.101.1",
            "roles": "",
            "iflabel": "",
            "wwpn": ""
        }
    },
]

ISCSI_PORT_PORTALS = {
    'p30': ['172.20.101.65:3260', '172.20.101.64:3260'],
    'p31': ['172.20.101.66:3260', '172.20.101.67:3260']
}

RES_MAPPING_RESP = [
    {
        "resource_id": "62bbb941-ba4a-4101-927d-e527ce5ee011",
        "host_id": "5e37d335-8fff-4aee-840a-34749301a16a",
        "mapped_lun_id": 1
    }
]

POOL_LIST_RESPONCE = [
    {
        "id": "7e259486-deb8-4d11-8cb0-e2c5874aaa5e",
        "name": "cinder-pool",
        "status": "ready"
    }
]

OK_POOL_ID = '7e259486-deb8-4d11-8cb0-e2c5874aaa5e'

WRONG_POOL_ID = 'wrong-id'

ERROR_VOLUME = [
    {
        "ptyId": "f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "id": "f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "name": "cinder-volume-f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "type": "block",
        "poolId": "92c05782-7529-479f-8db7-b9435e1e9a3d",
        "size": 16106127360,
        "maxModifySize": 95330557231104,
        "status": "error",
    }
]

READY_VOLUME = [
    {
        "ptyId": "f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "id": "f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "name": "cinder-volume-f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "type": "block",
        "poolId": "92c05782-7529-479f-8db7-b9435e1e9a3d",
        "size": 16106127360,
        "maxModifySize": 95330557231104,
        "status": "ready",
    }
]

ONLINE_VOLUME = [
    {
        "ptyId": "f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "id": "f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "name": "cinder-volume-f28ee814-22ed-4bb0-8b6a-f7ce9075034a",
        "type": "block",
        "poolId": "92c05782-7529-479f-8db7-b9435e1e9a3d",
        "size": 16106127360,
        "maxModifySize": 95330557231104,
        "status": "online",
    }
]


RESOURCE_INFORMATION = {
    "ptyId": "62bbb941-ba4a-4101-927d-e527ce5ee011",
    "id": "62bbb941-ba4a-4101-927d-e527ce5ee011",
    "name": "res1",
    "type": "block",
    "poolId": "c46584c5-3113-4cc7-8a72-f9262f32c508",
    "size": 1073741824,
    "maxModifySize": 5761094647808,
    "status": "ready",
    "stat": {
        "used_capacity": 1073741824,
        "mapped_blocks": 0,
        "dedup_count": 0,
        "reduction_ratio": 0
    },
    "lbaFormat": "4kn",
    "volume_id": "pty-vol-62bbb941-ba4a-4101-927d-e527ce5ee011",
    "wwid": "naa.614529011650000c4000800000000004",
    "lun_id": "4",
    "cached": "true",
    "rCacheMode": "enabled",
    "wCacheMode": "enabled",
    "ports": [
        {
            "port": "fc21",
            "port_status": "healthy",
            "port_status_desc":
            "resource is available on all storage controllers",
            "running": [
                "sp-1",
                "sp-0"
            ],
            "wwn": [
                "10:00:14:52:90:00:03:91",
                "10:00:14:52:90:00:03:11"
            ],
            "lun": "scsi-lun-fc21-4",
            "volume": "pty-vol-62bbb941-ba4a-4101-927d-e527ce5ee011",
            "lun_index": "4"
        },
        {
            "port": "fc20",
            "port_status": "healthy",
            "port_status_desc":
            "resource is available on all storage controllers",
            "running": [
                "sp-1",
                "sp-0"
            ],
            "wwn": [
                "10:00:14:52:90:00:03:10",
                "10:00:14:52:90:00:03:90"
            ],
            "lun": "scsi-lun-fc20-4",
            "volume": "pty-vol-62bbb941-ba4a-4101-927d-e527ce5ee011",
            "lun_index": "4"
        }
    ],
    "volume_path": "/dev/mapper/dmc-89382c6c-7cf9-4ff8-bdbb-f438d20c960a",
    "blockSize": "4kn",
    "replication": {
        "is_enabled": False
    }
}

POOL_NAME = 'cinder-pool'


class MockResponse:
    def __init__(self, json_data, status_code):
        self.json_data = json_data
        self.status_code = status_code

    def json(self):
        return self.json_data


class DummyVolume(object):
    def __init__(self, volid, volsize=1):
        self.id = volid
        self._name_id = None
        self.size = volsize
        self.status = None
        self.__volume_type_id = 1
        self.attach_status = None
        self.volume_attachment = None
        self.provider_location = None
        self.name = None
        self.metadata = {}

    @property
    def name_id(self):
        return self.id if not self._name_id else self._name_id

    @property
    def name(self):
        return self.name_id

    @property
    def volume_type_id(self):
        return self.__volume_type_id

    @name_id.setter
    def name_id(self, value):
        self._name_id = value

    @name.setter
    def name(self, value):
        self._name_id = value

    @volume_type_id.setter
    def volume_type_id(self, value):
        self.__volume_type_id = value


def get_fake_tatlin_config():
    config = configuration.Configuration(
        tatlin_opts,
        configuration.SHARED_CONF_GROUP)
    config.san_ip = '127.0.0.1'
    config.san_password = 'pwd'
    config.san_login = 'admin'
    config.pool_name = POOL_NAME
    config.host_group = 'cinder-group'
    config.tat_api_retry_count = 1
    config.wait_interval = 1
    config.wait_retry_count = 3
    config.chap_username = 'chap_user'
    config.chap_password = 'chap_passwd'
    config.state_path = '/tmp'
    return config


class TatlinCommonVolumeDriverTest(TestCase):
    @mock.patch.object(TatlinVolumeConnections, 'create_store')
    @mock.patch.object(TatlinAccessAPI, '_authenticate_access')
    def setUp(self, auth_access, create_store):
        access_api = TatlinAccessAPI('127.0.0.1', '443',
                                     'user', 'passwd', False)
        access_api._authenticate_access = MagicMock()
        self.client = TatlinClientCommon(access_api,
                                         api_retry_count=1,
                                         wait_interval=1,
                                         wait_retry_count=3)
        self.driver = TatlinCommonVolumeDriver(
            configuration=get_fake_tatlin_config())
        self.driver._get_tatlin_client = MagicMock()
        self.driver._get_tatlin_client.return_value = self.client
        self.driver.do_setup(None)

    @mock.patch.object(TatlinClientCommon, 'delete_volume')
    @mock.patch.object(TatlinClientCommon, 'is_volume_exists')
    def test_delete_volume_ok(self, is_volume_exist, delete_volume):
        cinder_manage.cfg.CONF.set_override('lock_path', '/tmp/locks',
                                            group='oslo_concurrency')
        is_volume_exist.side_effect = [True, False, False]
        self.driver.delete_volume(DummyVolume(VOLUME_NAME))

    @mock.patch.object(TatlinClientCommon, 'delete_volume')
    @mock.patch.object(TatlinClientCommon, 'is_volume_exists')
    def test_delete_volume_ok_404(self, is_volume_exist, delete_volume):
        cinder_manage.cfg.CONF.set_override('lock_path', '/tmp/locks',
                                            group='oslo_concurrency')
        is_volume_exist.side_effect = [False]
        self.driver.delete_volume(DummyVolume(VOLUME_NAME))

    @mock.patch.object(TatlinClientCommon, 'delete_volume')
    @mock.patch.object(TatlinClientCommon, 'is_volume_exists')
    def test_delete_volume_error_500(self, is_volume_exist, delete_volume):
        cinder_manage.cfg.CONF.set_override('lock_path', '/tmp/locks',
                                            group='oslo_concurrency')
        is_volume_exist.return_value = True
        delete_volume.side_effect = TatlinAPIException(500, 'ERROR')
        with self.assertRaises(VolumeBackendAPIException):
            self.driver.delete_volume(DummyVolume(VOLUME_NAME))

    @mock.patch.object(TatlinCommonVolumeDriver, '_update_qos')
    @mock.patch.object(TatlinClientCommon, 'is_volume_ready')
    @mock.patch.object(TatlinClientCommon, 'extend_volume')
    @mock.patch.object(TatlinClientCommon, 'is_volume_exists')
    def test_extend_volume_ok(self,
                              is_volume_exist,
                              extend_volume,
                              is_volume_ready,
                              update_qos):
        cinder_manage.cfg.CONF.set_override('lock_path', '/tmp/locks',
                                            group='oslo_concurrency')
        is_volume_ready.return_value = True
        is_volume_exist.return_value = True
        self.driver.extend_volume(DummyVolume(VOLUME_NAME), 10)

    @mock.patch('time.sleep')
    @mock.patch.object(TatlinCommonVolumeDriver, '_update_qos')
    @mock.patch.object(TatlinClientCommon, 'is_volume_ready')
    @mock.patch.object(TatlinClientCommon, 'extend_volume')
    @mock.patch.object(TatlinClientCommon, 'is_volume_exists')
    def test_extend_volume_error_not_ready(self,
                                           is_volume_exist,
                                           extend_volume,
                                           is_volume_ready,
                                           update_qos,
                                           sleeper):
        cinder_manage.cfg.CONF.set_override('lock_path', '/tmp/locks',
                                            group='oslo_concurrency')
        is_volume_ready.return_value = False
        is_volume_exist.return_value = True
        with self.assertRaises(VolumeBackendAPIException):
            self.driver.extend_volume(DummyVolume(VOLUME_NAME), 10)

    @mock.patch.object(TatlinClientCommon,
                       'is_volume_ready',
                       return_value=True)
    def test_wait_volume_reay_success(self, is_ready):
        self.driver.wait_volume_ready(DummyVolume('cinder_volume'))

    @mock.patch.object(TatlinCommonVolumeDriver, '_update_qos')
    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_succeess_manage_existing(self, sendMock, qosMock):
        sendMock.side_effect = [
            (MockResponse([{'id': '1', 'poolId': OK_POOL_ID}], 200)),
            (MockResponse(POOL_LIST_RESPONCE, 200))
        ]
        self.driver.manage_existing(DummyVolume(VOLUME_NAME), {
            'source-name': 'existing-resource'
        })

    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_fail_manage_existing_volume_not_found(self, sendMock):
        self.driver.tatlin_api._send_request = Mock()
        sendMock.side_effect = [
            (MockResponse([{}], 404)),
        ]

        with self.assertRaises(ManageExistingInvalidReference):
            self.driver.manage_existing(DummyVolume('new-vol-id'), {
                'source-name': 'existing-resource'
            })
            self.driver.tatlin_api.get_volume_info.assert_called_once()
            self.driver.tatlin_api.get_pool_id_by_name.assert_not_called()

    @mock.patch.object(TatlinCommonVolumeDriver, '_update_qos')
    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_fail_manage_existing_wrong_pool(self, sendMock, qosMock):
        sendMock.side_effect = [
            (MockResponse([{'id': '1', 'poolId': WRONG_POOL_ID}], 200)),
            (MockResponse(POOL_LIST_RESPONCE, 200))
        ]

        with self.assertRaises(ManageExistingInvalidReference):
            self.driver.manage_existing(DummyVolume('new-vol-id'), {
                'source-name': 'existing-resource'
            })
            self.driver.tatlin_api.get_volume_info.assert_called_once()
            self.driver.tatlin_api.get_pool_id_by_name.assert_called_once()

    @mock.patch.object(TatlinClientCommon, 'get_resource_count')
    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_success_create_volume(self, send_requst, object_count):
        self.driver._stats['overall_resource_count'] = 1
        object_count.side_effect = [(1, 1)]
        send_requst.side_effect = [
            (MockResponse(POOL_LIST_RESPONCE, 200)),      # Get pool id
            (MockResponse({}, 200)),                      # Create volume
            (MockResponse(READY_VOLUME, 200)),          # Is volume ready
            (MockResponse(READY_VOLUME, 200))           # Is volume ready
        ]
        self.driver._update_qos = Mock()
        self.driver.create_volume(DummyVolume(VOLUME_NAME))

    @mock.patch.object(TatlinClientCommon, 'get_resource_count')
    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_fail_create_volume_400(self, send_request, object_count):
        self.driver._stats['overall_resource_count'] = 1
        object_count.side_effect = [(1, 1)]
        send_request.side_effect = [
            (MockResponse(POOL_LIST_RESPONCE, 200)),
            (MockResponse({}, 500)),
            (MockResponse({}, 400))
        ]
        with self.assertRaises(VolumeBackendAPIException):
            self.driver.create_volume(DummyVolume(VOLUME_NAME))
            self.driver.tatlin_api.create_volume.assert_called_once()

    @mock.patch('time.sleep')
    @mock.patch.object(TatlinClientCommon, 'get_resource_count')
    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_fail_volume_not_ready_create_volume(self, sendMock,
                                                 volume_count, sleeper):
        self.driver._stats['overall_resource_count'] = 1
        volume_count.side_effect = [(1, 1)]
        sendMock.side_effect = [
            (MockResponse(POOL_LIST_RESPONCE, 200)),
            (MockResponse({}, 200)),
            (MockResponse(ERROR_VOLUME, 200)),
            (MockResponse(ERROR_VOLUME, 200)),
            (MockResponse(ERROR_VOLUME, 200)),
        ]
        with self.assertRaises(VolumeBackendAPIException):
            self.driver.create_volume(DummyVolume(VOLUME_NAME))

    @mock.patch.object(TatlinCommonVolumeDriver, '_get_ports_portals')
    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_fail_create_export(self, sendMock, portsMock):
        sendMock.side_effect = [
            (MockResponse(OSMGR_ISCSI_PORTS, 200)),
        ]

        portsMock.side_effect = [
            ISCSI_PORT_PORTALS
        ]
        self.driver._is_all_ports_assigned = Mock(return_value=True)
        with self.assertRaises(NotImplementedError):
            self.driver.create_export(None, DummyVolume(VOLUME_NAME), None)

    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_find_mapped_lun(self, sendMock):
        sendMock.side_effect = [
            (MockResponse(RES_MAPPING_RESP, 200)),
        ]

        self.driver.find_current_host = Mock(
            return_value='5e37d335-8fff-4aee-840a-34749301a16a')
        self.driver._find_mapped_lun(
            '62bbb941-ba4a-4101-927d-e527ce5ee011', '')

    @mock.patch.object(TatlinCommonVolumeDriver, '_update_qos')
    @mock.patch.object(TatlinCommonVolumeDriver, 'wait_volume_online')
    @mock.patch.object(TatlinClientCommon, 'add_vol_to_host')
    @mock.patch.object(TatlinClientCommon,
                       'is_volume_exists',
                       return_value=True)
    @mock.patch.object(TatlinAccessAPI, 'send_request')
    def test_add_volume_to_host(self,
                                *args):
        vol = DummyVolume('62bbb941-ba4a-4101-927d-e527ce5ee011')
        self.driver.add_volume_to_host(
            vol, '5e37d335-8fff-4aee-840a-34749301a16a'
        )



# Copyright (c) 2021 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import errno
import os
from unittest import mock

import ddt
from oslo_concurrency import processutils as putils
from oslo_utils import imageutils
from oslo_utils import units

from cinder import context
from cinder import exception
from cinder.image import image_utils
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.volume import configuration as conf
from cinder.volume.drivers.dell_emc.powerstore import nfs
from cinder.volume import volume_utils


NFS_CONFIG = {'max_over_subscription_ratio': 1.0,
              'reserved_percentage': 0,
              'nfs_sparsed_volumes': True,
              'nfs_qcow2_volumes': False,
              'nas_secure_file_permissions': 'false',
              'nas_secure_file_operations': 'false'}


QEMU_IMG_INFO_OUT1 = """image: %(volid)s
        file format: raw
        virtual size: %(size_gb)sG (%(size_b)s bytes)
        disk size: 173K
        """

QEMU_IMG_INFO_OUT2 = """image: %(volid)s
        file format: qcow2
        virtual size: %(size_gb)sG (%(size_b)s bytes)
        disk size: 173K
        """

QEMU_IMG_INFO_OUT3 = """image: volume-%(volid)s.%(snapid)s
        file format: qcow2
        virtual size: %(size_gb)sG (%(size_b)s bytes)
        disk size: 196K
        cluster_size: 65536
        backing file: volume-%(volid)s
        backing file format: qcow2
        Format specific information:
            compat: 1.1
            lazy refcounts: false
            refcount bits: 16
            corrupt: false
        """


@ddt.ddt
class PowerStoreNFSDriverInitializeTestCase(test.TestCase):
    TEST_NFS_HOST = 'nfs-host1'

    def setUp(self):
        super(PowerStoreNFSDriverInitializeTestCase, self).setUp()
        self.context = mock.Mock()
        self.create_configuration()
        self.override_config('compute_api_class', 'unittest.mock.Mock')
        self.drv = nfs.PowerStoreNFSDriverInitialization(
            configuration=self.configuration)

    def create_configuration(self):
        config = conf.Configuration(None)
        config.append_config_values(nfs.nfs_opts)
        self.configuration = config

    def test_check_multiattach_support(self):
        drv = self.drv

        self.configuration.nfs_qcow2_volumes = False

        drv._check_multiattach_support()
        self.assertEqual(not self.configuration.nfs_qcow2_volumes,
                         drv.multiattach_support)

    def test_check_multiattach_support_disable(self):
        drv = self.drv
        drv.configuration.nfs_qcow2_volumes = True

        drv._check_multiattach_support()
        self.assertEqual(not self.configuration.nfs_qcow2_volumes,
                         drv.multiattach_support)

    def test_check_snapshot_support(self):
        drv = self.drv
        drv.configuration.nfs_snapshot_support = True
        drv.configuration.nas_secure_file_operations = 'false'

        drv._check_snapshot_support()

        self.assertTrue(drv.configuration.nfs_snapshot_support)

    def test_check_snapshot_support_disable(self):
        drv = self.drv
        drv.configuration.nfs_snapshot_support = False
        drv.configuration.nas_secure_file_operations = 'false'

        self.assertRaises(exception.VolumeDriverException,
                          drv._check_snapshot_support)

    def test_check_snapshot_support_nas_true(self):
        drv = self.drv
        drv.configuration.nfs_snapshot_support = True
        drv.configuration.nas_secure_file_operations = 'true'

        self.assertRaises(exception.VolumeDriverException,
                          drv._check_snapshot_support)

    @mock.patch("cinder.volume.drivers.nfs.NfsDriver.do_setup")
    def test_do_setup(self, mock_super_do_setup):
        drv = self.drv
        drv.configuration.nas_host = self.TEST_NFS_HOST

        mock_check_multiattach_support = self.mock_object(
            drv, '_check_multiattach_support'
        )

        drv.do_setup(self.context)

        self.assertTrue(mock_check_multiattach_support.called)

    def test_check_package_is_installed(self):
        drv = self.drv
        package = 'dellfcopy'
        mock_execute = self.mock_object(drv, '_execute')
        drv._check_package_is_installed(package)
        mock_execute.assert_called_once_with(package,
                                             check_exit_code=False,
                                             run_as_root=False)

    def test_check_package_is_not_installed(self):
        drv = self.drv
        package = 'dellfcopy'
        drv._execute = mock.Mock(
            side_effect=OSError(
                errno.ENOENT, 'No such file or directory'
            )
        )
        self.assertRaises(exception.VolumeDriverException,
                          drv._check_package_is_installed, package)
        drv._execute.assert_called_once_with(package,
                                             check_exit_code=False,
                                             run_as_root=False)

    def test_check_for_setup_error(self):
        drv = self.drv
        mock_check_package_is_installed = self.mock_object(
            drv, '_check_package_is_installed')
        drv.check_for_setup_error()
        mock_check_package_is_installed.assert_called_once_with('dellfcopy')

    def test_check_for_setup_error_not_passed(self):
        drv = self.drv
        drv._execute = mock.Mock(
            side_effect=OSError(
                errno.ENOENT, 'No such file or directory'
            )
        )
        self.assertRaises(exception.VolumeDriverException,
                          drv.check_for_setup_error)
        drv._execute.assert_called_once_with('dellfcopy',
                                             check_exit_code=False,
                                             run_as_root=False)

    def test_update_volume_stats_has_multiattach(self):
        drv = self.drv

        self.mock_object(nfs.NfsDriver, '_update_volume_stats')
        drv.multiattach_support = True
        drv._stats = {}

        drv._update_volume_stats()

        self.assertIn('multiattach', drv._stats)
        self.assertTrue(drv._stats['multiattach'])


@ddt.ddt
class PowerStoreNFSDriverTestCase(test.TestCase):
    TEST_NFS_HOST = 'nfs-host1'
    TEST_NFS_SHARE_PATH = '/export'
    TEST_NFS_EXPORT = '%s:%s' % (TEST_NFS_HOST, TEST_NFS_SHARE_PATH)
    TEST_SIZE_IN_GB = 1
    TEST_MNT_POINT = '/mnt/nfs'
    TEST_MNT_POINT_BASE_EXTRA_SLASH = '/opt/stack/data/cinder//mnt'
    TEST_MNT_POINT_BASE = '/mnt/test'
    TEST_LOCAL_PATH = '/mnt/nfs/volume-123'
    TEST_FILE_NAME = 'test.txt'
    VOLUME_UUID = 'abcdefab-cdef-abcd-efab-cdefabcdefab'

    def setUp(self):
        super(PowerStoreNFSDriverTestCase, self).setUp()

        self.configuration = mock.Mock(conf.Configuration)
        self.configuration.append_config_values(mock.ANY)
        self.configuration.nfs_sparsed_volumes = True
        self.configuration.nas_secure_file_permissions = 'false'
        self.configuration.nas_secure_file_operations = 'false'
        self.configuration.nfs_mount_point_base = self.TEST_MNT_POINT_BASE
        self.configuration.nfs_snapshot_support = True
        self.configuration.max_over_subscription_ratio = 1.0
        self.configuration.reserved_percentage = 5
        self.configuration.nfs_mount_options = None
        self.configuration.nfs_qcow2_volumes = True
        self.configuration.nas_host = '0.0.0.0'
        self.configuration.nas_share_path = None

        self.mock_object(volume_utils, 'get_max_over_subscription_ratio',
                         return_value=1)
        self.context = context.get_admin_context()

        self._driver = nfs.PowerStoreNFSDriver(
            configuration=self.configuration)
        self._driver.shares = {}
        self.mock_object(self._driver, '_execute')

    def test_do_fast_clone_file(self):
        drv = self._driver
        volume_path = 'fake/path'
        new_volume_path = 'fake/new_path'

        drv._do_fast_clone_file(volume_path, new_volume_path)

        drv._execute.assert_called_once_with(
            'dellfcopy', '-o', 'fastclone', '-s', volume_path, '-d',
            new_volume_path, '-v', '1', run_as_root=True
        )

    def test_do_fast_clone_file_raise_error(self):
        drv = self._driver
        volume_path = 'fake/path'
        new_volume_path = 'fake/new_path'

        drv._execute = mock.Mock(
            side_effect=putils.ProcessExecutionError()
        )
        self.assertRaises(putils.ProcessExecutionError,
                          drv._do_fast_clone_file, volume_path,
                          new_volume_path)
        drv._execute.assert_called_once_with(
            'dellfcopy', '-o', 'fastclone', '-s', volume_path, '-d',
            new_volume_path, '-v', '1', run_as_root=True
        )

    def _simple_volume(self, **kwargs):
        updates = {'id': self.VOLUME_UUID,
                   'provider_location': self.TEST_NFS_EXPORT,
                   'display_name': f'volume-{self.VOLUME_UUID}',
                   'name': f'volume-{self.VOLUME_UUID}',
                   'size': 10,
                   'status': 'available'}

        updates.update(kwargs)
        if 'display_name' not in updates:
            updates['display_name'] = 'volume-%s' % updates['id']

        return fake_volume.fake_volume_obj(self.context, **updates)

    def test_delete_volume_without_info(self):
        drv = self._driver
        volume = fake_volume.fake_volume_obj(
            self.context,
            display_name='volume',
            provider_location=self.TEST_NFS_EXPORT
        )
        vol_path = '/path/to/vol'

        mock_ensure_share_mounted = self.mock_object(
            drv, '_ensure_share_mounted')
        mock_local_path_volume_info = self.mock_object(
            drv, '_local_path_volume_info'
        )
        mock_local_path_volume_info.return_value = self.TEST_LOCAL_PATH
        mock_read_info_file = self.mock_object(drv, '_read_info_file')
        mock_read_info_file.return_value = {}
        mock_local_path_volume = self.mock_object(drv, '_local_path_volume')
        mock_local_path_volume.return_value = vol_path

        drv.delete_volume(volume)

        mock_ensure_share_mounted.assert_called_once_with(
            self.TEST_NFS_EXPORT)
        mock_local_path_volume.assert_called_once_with(volume)
        mock_read_info_file.assert_called_once_with(
            self.TEST_LOCAL_PATH, empty_if_missing=True)
        mock_local_path_volume.assert_called_once_with(volume)
        drv._execute.assert_called_once_with(
            'rm', '-f', vol_path, run_as_root=True)

    def test_delete_volume_with_info(self):
        drv = self._driver
        volume = fake_volume.fake_volume_obj(
            self.context,
            display_name='volume',
            provider_location=self.TEST_NFS_EXPORT
        )
        vol_path = '/path/to/vol'
        with mock.patch.object(drv, '_ensure_share_mounted'):
            mock_local_path_volume_info = self.mock_object(
                drv, '_local_path_volume_info'
            )
            mock_local_path_volume_info.return_value = self.TEST_LOCAL_PATH
            mock_read_info_file = self.mock_object(drv, '_read_info_file')
            mock_read_info_file.return_value = {'active': '/path/to/active'}
            mock_local_path_volume = self.mock_object(
                drv, '_local_path_volume')
            mock_local_path_volume.return_value = vol_path

            drv.delete_volume(volume)

        self.assertEqual(drv._execute.call_count, 3)

    def test_delete_volume_without_provider_location(self):
        drv = self._driver
        volume = fake_volume.fake_volume_obj(
            self.context,
            display_name='volume',
            provider_location=''
        )
        drv.delete_volume(volume)

        self.assertFalse(bool(drv._execute.call_count))

    @ddt.data([None, QEMU_IMG_INFO_OUT1],
              ['raw', QEMU_IMG_INFO_OUT1],
              ['qcow2', QEMU_IMG_INFO_OUT2])
    @ddt.unpack
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    def test_extend_volume(self, file_format, qemu_img_info, mock_get):
        drv = self._driver
        volume = fake_volume.fake_volume_obj(
            self.context,
            id='80ee16b6-75d2-4d54-9539-ffc1b4b0fb10',
            size=1,
            provider_location='nfs_share')
        if file_format:
            volume.admin_metadata = {'format': file_format}
        mock_get.return_value = volume
        path = 'path'
        new_size = volume['size'] + 1

        mock_img_utils = self.mock_object(drv, '_qemu_img_info')
        img_out = qemu_img_info % {'volid': volume.id,
                                   'size_gb': volume.size,
                                   'size_b': volume.size * units.Gi}
        mock_img_utils.return_value = imageutils.QemuImgInfo(
            img_out)

        with mock.patch.object(image_utils, 'resize_image') as resize:
            with mock.patch.object(drv, 'local_path', return_value=path):
                with mock.patch.object(drv, '_is_share_eligible',
                                       return_value=True):

                    drv.extend_volume(volume, new_size)

                    resize.assert_called_once_with(path, new_size)

    def test_create_volume_from_snapshot(self):
        drv = self._driver
        src_volume = self._simple_volume(size=10)
        src_volume.id = fake.VOLUME_ID

        fake_snap = fake_snapshot.fake_snapshot_obj(self.context)
        fake_snap.volume = src_volume
        fake_snap.size = 10
        fake_snap.status = 'available'

        new_volume = self._simple_volume(size=src_volume.size)

        drv._find_share = mock.Mock(return_value=self.TEST_NFS_EXPORT)
        drv._copy_volume_from_snapshot = mock.Mock()

        drv._create_volume_from_snapshot(new_volume, fake_snap)

        drv._find_share.assert_called_once_with(new_volume)
        drv._copy_volume_from_snapshot.assert_called_once_with(
            fake_snap, new_volume, new_volume.size
        )

    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    def test_create_cloned_volume(self, mock_get):
        drv = self._driver
        volume = self._simple_volume()
        mock_get.return_value = volume
        vol_dir = os.path.join(self.TEST_MNT_POINT_BASE,
                               drv._get_hash_str(volume.provider_location))
        vol_path = os.path.join(vol_dir, volume.name)

        new_volume = self._simple_volume()
        new_vol_dir = os.path.join(self.TEST_MNT_POINT_BASE,
                                   drv._get_hash_str(
                                       volume.provider_location))
        new_vol_path = os.path.join(new_vol_dir, volume.name)

        drv._create_cloned_volume(new_volume, volume, self.context)

        command = ['dellfcopy', '-o', 'fastclone', '-s', vol_path,
                   '-d', new_vol_path, '-v', '1']
        calls = [mock.call(*command, run_as_root=True)]
        drv._execute.assert_has_calls(calls)

    @ddt.data([QEMU_IMG_INFO_OUT3])
    @ddt.unpack
    @mock.patch('cinder.objects.volume.Volume.save')
    def test_copy_volume_from_snapshot(self, qemu_img_info, mock_save):
        drv = self._driver
        src_volume = self._simple_volume(size=10)
        src_volume.id = fake.VOLUME_ID

        fake_snap = fake_snapshot.fake_snapshot_obj(self.context)
        snap_file = src_volume.name + '.' + fake_snap.id
        fake_snap.volume = src_volume
        fake_snap.size = 10
        fake_source_vol_path = os.path.join(
            drv._local_volume_dir(fake_snap.volume),
            src_volume.name
        )

        new_volume = self._simple_volume(size=10)
        new_vol_dir = os.path.join(self.TEST_MNT_POINT_BASE,
                                   drv._get_hash_str(
                                       src_volume.provider_location))
        new_vol_path = os.path.join(new_vol_dir, new_volume.name)

        mock_read_info_file = self.mock_object(drv, '_read_info_file')
        mock_read_info_file.return_value = {'active': snap_file,
                                            fake_snap.id: snap_file}

        mock_img_utils = self.mock_object(drv, '_qemu_img_info')
        img_out = qemu_img_info % {'volid': src_volume.id,
                                   'snapid': fake_snap.id,
                                   'size_gb': src_volume.size,
                                   'size_b': src_volume.size * units.Gi}
        mock_img_utils.return_value = imageutils.QemuImgInfo(img_out)

        drv._copy_volume_from_snapshot(fake_snap, new_volume, new_volume.size)

        command = ['dellfcopy', '-o', 'fastclone', '-s', fake_source_vol_path,
                   '-d', new_vol_path, '-v', '1']
        calls = [mock.call(*command, run_as_root=True)]
        drv._execute.assert_has_calls(calls)



#!/usr/bin/env python
# Copyright (c) 2011 X.commerce, a business unit of eBay Inc.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Interactive shell based on Django:
#
# Copyright (c) 2005, the Lawrence Journal-World
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#
#     3. Neither the name of Django nor the names of its contributors may be
#        used to endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""CLI interface for cinder management."""

import collections
import collections.abc as collections_abc
import errno
import glob
import itertools
import logging as python_logging
import os
import re
import sys
import time

from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import log as logging
from oslo_utils import timeutils
import tabulate

# Need to register global_opts
from cinder.backup import rpcapi as backup_rpcapi
from cinder.common import config  # noqa
from cinder import context
from cinder import db
from cinder.db import migration as db_migration
from cinder.db.sqlalchemy import api as db_api
from cinder.db.sqlalchemy import models
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import base as ovo_base
from cinder import quota
from cinder import rpc
from cinder.scheduler import rpcapi as scheduler_rpcapi
from cinder import version
from cinder.volume import rpcapi as volume_rpcapi
from cinder.volume import volume_utils


CONF = cfg.CONF

LOG = logging.getLogger(__name__)

RPC_VERSIONS = {
    'cinder-scheduler': scheduler_rpcapi.SchedulerAPI.RPC_API_VERSION,
    'cinder-volume': volume_rpcapi.VolumeAPI.RPC_API_VERSION,
    'cinder-backup': backup_rpcapi.BackupAPI.RPC_API_VERSION,
}

OVO_VERSION = ovo_base.OBJ_VERSIONS.get_current()


# Decorators for actions
def args(*args, **kwargs):
    args = list(args)
    if not args[0].startswith('-') and '-' in args[0]:
        kwargs.setdefault('metavar', args[0])
        args[0] = args[0].replace('-', '_')

    def _decorator(func):
        func.__dict__.setdefault('args', []).insert(0, (args, kwargs))
        return func
    return _decorator


class HostCommands(object):
    """List hosts."""

    @args('zone', nargs='?', default=None,
          help='Availability Zone (default: %(default)s)')
    def list(self, zone=None):
        """Show a list of all physical hosts.

        Can be filtered by zone.
        args: [zone]
        """
        print(_("%(host)-25s\t%(zone)-15s") % {'host': 'host', 'zone': 'zone'})
        ctxt = context.get_admin_context()
        services = objects.ServiceList.get_all(ctxt)
        if zone:
            services = [s for s in services if s.availability_zone == zone]
        hosts = []
        for srv in services:
            if not [h for h in hosts if h['host'] == srv['host']]:
                hosts.append(srv)

        for h in hosts:
            print(_("%(host)-25s\t%(availability_zone)-15s")
                  % {'host': h['host'],
                     'availability_zone': h['availability_zone']})


class DbCommands(object):
    """Class for managing the database."""

    # NOTE: Online migrations cannot depend on having Cinder services running.
    # Migrations can be called during Fast-Forward Upgrades without having any
    # Cinder services up.
    # NOTE: Online migrations must be removed at the beginning of the next
    # release to the one they've been introduced.  A comment with the release
    # a migration is introduced and the one where it must be removed must
    # preceed any element of the "online_migrations" tuple, like this:
    #    # Added in Queens remove in Rocky
    #    db.service_uuids_online_data_migration,
    online_migrations = (
        # TODO: (Z Release) Remove next line and this comment
        # TODO: (Y Release) Uncomment next line and remove this comment
        # db.remove_temporary_admin_metadata_data_migration,

        # TODO: (Y Release) Remove next 2 line and this comment
        db.volume_use_quota_online_data_migration,
        db.snapshot_use_quota_online_data_migration,
    )

    def __init__(self):
        pass

    @args('version', nargs='?', default=None, type=int,
          help='Database version')
    @args('--bump-versions', dest='bump_versions', default=False,
          action='store_true',
          help='Update RPC and Objects versions when doing offline upgrades, '
               'with this we no longer need to restart the services twice '
               'after the upgrade to prevent ServiceTooOld exceptions.')
    def sync(self, version=None, bump_versions=False):
        """Sync the database up to the most recent version."""
        if version is not None and version > db.MAX_INT:
            print(_('Version should be less than or equal to '
                    '%(max_version)d.') % {'max_version': db.MAX_INT})
            sys.exit(1)
        try:
            db_migration.db_sync(version)
        except db_exc.DBMigrationError as ex:
            print("Error during database migration: %s" % ex)
            sys.exit(1)

        try:
            if bump_versions:
                ctxt = context.get_admin_context()
                services = objects.ServiceList.get_all(ctxt)
                for service in services:
                    rpc_version = RPC_VERSIONS[service.binary]
                    if (service.rpc_current_version != rpc_version or
                            service.object_current_version != OVO_VERSION):
                        service.rpc_current_version = rpc_version
                        service.object_current_version = OVO_VERSION
                        service.save()
        except Exception as ex:
            print(_('Error during service version bump: %s') % ex)
            sys.exit(2)

    def version(self):
        """Print the current database version."""
        print(db_migration.db_version())

    @args('age_in_days', type=int,
          help='Purge deleted rows older than age in days')
    def purge(self, age_in_days):
        """Purge deleted rows older than a given age from cinder tables."""
        age_in_days = int(age_in_days)
        if age_in_days < 0:
            print(_("Must supply a positive value for age"))
            sys.exit(1)
        if age_in_days >= (int(time.time()) / 86400):
            print(_("Maximum age is count of days since epoch."))
            sys.exit(1)
        ctxt = context.get_admin_context()

        try:
            db.purge_deleted_rows(ctxt, age_in_days)
        except db_exc.DBReferenceError:
            print(_("Purge command failed, check cinder-manage "
                    "logs for more details."))
            sys.exit(1)

    def _run_migration(self, ctxt, max_count):
        ran = 0
        exceptions = False
        migrations = {}
        for migration_meth in self.online_migrations:
            count = max_count - ran
            try:
                found, done = migration_meth(ctxt, count)
            except Exception:
                msg = (_("Error attempting to run %(method)s") %
                       {'method': migration_meth.__name__})
                print(msg)
                LOG.exception(msg)
                exceptions = True
                found = done = 0

            name = migration_meth.__name__
            if found:
                print(_('%(found)i rows matched query %(meth)s, %(done)i '
                        'migrated') % {'found': found,
                                       'meth': name,
                                       'done': done})
            migrations[name] = found, done
            if max_count is not None:
                ran += done
                if ran >= max_count:
                    break
        return migrations, exceptions

    @args('--max_count', metavar='<number>', dest='max_count', type=int,
          help='Maximum number of objects to consider.')
    def online_data_migrations(self, max_count=None):
        """Perform online data migrations for the release in batches."""
        ctxt = context.get_admin_context()
        if max_count is not None:
            unlimited = False
            if max_count < 1:
                print(_('Must supply a positive value for max_count.'))
                sys.exit(127)
        else:
            unlimited = True
            max_count = 50
            print(_('Running batches of %i until complete.') % max_count)

        ran = None
        exceptions = False
        migration_info = {}
        while ran is None or ran != 0:
            migrations, exceptions = self._run_migration(ctxt, max_count)
            ran = 0
            for name in migrations:
                migration_info.setdefault(name, (0, 0))
                migration_info[name] = (
                    max(migration_info[name][0], migrations[name][0]),
                    migration_info[name][1] + migrations[name][1],
                )
                ran += migrations[name][1]
            if not unlimited:
                break
        headers = ["{}".format(_('Migration')),
                   "{}".format(_('Total Needed')),
                   "{}".format(_('Completed')), ]
        rows = []
        for name in sorted(migration_info.keys()):
            info = migration_info[name]
            rows.append([name, info[0], info[1]])
        print(tabulate.tabulate(rows, headers=headers, tablefmt='psql'))

        # NOTE(imacdonn): In the "unlimited" case, the loop above will only
        # terminate when all possible migrations have been effected. If we're
        # still getting exceptions, there's a problem that requires
        # intervention. In the max-count case, exceptions are only considered
        # fatal if no work was done by any other migrations ("not ran"),
        # because otherwise work may still remain to be done, and that work
        # may resolve dependencies for the failing migrations.
        if exceptions and (unlimited or not ran):
            print(_("Some migrations failed unexpectedly. Check log for "
                    "details."))
            sys.exit(2)

        sys.exit(1 if ran else 0)

    @args('--enable-replication', action='store_true', default=False,
          help='Set replication status to enabled (default: %(default)s).')
    @args('--active-backend-id', default=None,
          help='Change the active backend ID (default: %(default)s).')
    @args('--backend-host', required=True,
          help='The backend host name.')
    def reset_active_backend(self, enable_replication, active_backend_id,
                             backend_host):
        """Reset the active backend for a host."""

        ctxt = context.get_admin_context()

        try:
            db.reset_active_backend(ctxt, enable_replication,
                                    active_backend_id, backend_host)
        except db_exc.DBReferenceError:
            print(_("Failed to reset active backend for host %s, "
                    "check cinder-manage logs for more details.") %
                  backend_host)
            sys.exit(1)


class QuotaCommands(object):
    """Class for managing quota issues."""

    def __init__(self):
        pass

    @args('--project-id', default=None,
          help=('The ID of the project where we want to sync the quotas '
                '(defaults to all projects).'))
    def check(self, project_id):
        """Check if quotas and reservations are correct

        This action checks quotas and reservations, for a specific project or
        for all projects, to see if they are out of sync.

        The check will also look for duplicated entries.

        One way to use this check in combination with the sync action is to
        run the check for all projects, take note of those that are out of
        sync, and then sync them one by one at intervals to reduce stress on
        the DB.
        """
        result = self._check_sync(project_id, do_fix=False)
        if result:
            sys.exit(1)

    @args('--project-id', default=None,
          help=('The ID of the project where we want to sync the quotas '
                '(defaults to all projects).'))
    def sync(self, project_id):
        """Fix quotas and reservations

        This action refreshes existing quota usage and reservation count for a
        specific project or for all projects.

        The refresh will also remove duplicated entries.

        This operation is best executed when Cinder is not running, but it can
        be run with cinder services running as well.

        A different transaction is used for each project's quota sync, so an
        action failure will only rollback the current project's changes.
        """
        self._check_sync(project_id, do_fix=True)

    @db_api.main_context_manager.reader
    def _get_quota_projects(self, context, project_id):
        """Get project ids that have quota_usage entries."""
        if project_id:
            model = models.QuotaUsage
            # If the project does not exist
            if not context.session.query(
                db_api.sql.exists()
                .where(
                    db_api.and_(
                        model.project_id == project_id,
                        ~model.deleted,
                    ),
                )
            ).scalar():
                print(
                    'Project id %s has no quota usage. Nothing to do.' %
                    project_id,
                )
                return []
            return [project_id]

        projects = db_api.model_query(
            context,
            models.QuotaUsage,
            read_deleted="no"
        ).with_entities('project_id').distinct().all()
        project_ids = [row.project_id for row in projects]
        return project_ids

    def _get_usages(self, context, resources, project_id):
        """Get data necessary to check out of sync quota usage.

        Returns a list QuotaUsage instances for the specific project
        """
        usages = db_api.model_query(
            context,
            db_api.models.QuotaUsage,
            read_deleted="no",
        ).filter_by(project_id=project_id).with_for_update().all()
        return usages

    def _get_reservations(self, context, project_id, usage_id):
        """Get reservations for a given project and usage id."""
        reservations = (
            db_api.model_query(
                context,
                models.Reservation,
                read_deleted="no",
            )
            .filter_by(project_id=project_id, usage_id=usage_id)
            .with_for_update()
            .all()
        )
        return reservations

    def _check_duplicates(self, context, usages, do_fix):
        """Look for duplicated quota used entries (bug#1484343)

        If we have duplicates and we are fixing them, then we reassign the
        reservations of the usage we are removing.
        """
        resources = collections.defaultdict(list)
        for usage in usages:
            resources[usage.resource].append(usage)

        duplicates_found = False
        result = []
        for resource_usages in resources.values():
            keep_usage = resource_usages[0]
            if len(resource_usages) > 1:
                duplicates_found = True
                print('\t%s: %s duplicated usage entries - ' %
                      (keep_usage.resource, len(resource_usages) - 1),
                      end='')

                if do_fix:
                    # Each of the duplicates can have reservations
                    reassigned = 0
                    for usage in resource_usages[1:]:
                        reservations = self._get_reservations(
                            context,
                            usage.project_id,
                            usage.id,
                        )
                        reassigned += len(reservations)
                        for reservation in reservations:
                            reservation.usage_id = keep_usage.id
                        keep_usage.in_use += usage.in_use
                        keep_usage.reserved += usage.reserved
                        usage.delete(context.session)
                    print('duplicates removed & %s reservations reassigned' %
                          reassigned)
                else:
                    print('ignored')
            result.append(keep_usage)
        return result, duplicates_found

    def _check_sync(self, project_id, do_fix):
        """Check the quotas and reservations optionally fixing them."""

        ctxt = context.get_admin_context()
        # Get the quota usage types and their sync methods
        resources = quota.QUOTAS.resources

        # Get all project ids that have quota usage. Method doesn't lock
        # projects, since newly added projects should not be out of sync and
        # projects removed will just turn nothing on the quota usage.
        projects = self._get_quota_projects(ctxt, project_id)

        discrepancy = False

        for project in projects:
            discrepancy &= self._check_project_sync(
                ctxt,
                project,
                do_fix,
                resources,
            )

        print('Action successfully completed')
        return discrepancy

    @db_api.main_context_manager.reader
    def _check_project_sync(self, context, project, do_fix, resources):
        print('Processing quota usage for project %s' % project)

        discrepancy = False
        action_msg = ' - fixed' if do_fix else ''

        # NOTE: It's important to always get the quota first and then the
        # reservations to prevent deadlocks with quota commit and rollback from
        # running Cinder services.

        # We only want to sync existing quota usage rows
        usages = self._get_usages(context, resources, project)

        # Check for duplicated entries (bug#1484343)
        usages, duplicates_found = self._check_duplicates(
            context, usages, do_fix,
        )
        if duplicates_found:
            discrepancy = True

        # Check quota and reservations
        for usage in usages:
            resource_name = usage.resource
            # Get the correct value for this quota usage resource
            updates = db_api._get_sync_updates(
                context,
                project,
                resources,
                resource_name,
            )
            in_use = updates[resource_name]
            if in_use != usage.in_use:
                print(
                    '\t%s: invalid usage saved=%s actual=%s%s' %
                    (resource_name, usage.in_use, in_use, action_msg)
                )
                discrepancy = True
                if do_fix:
                    usage.in_use = in_use

            reservations = self._get_reservations(
                context,
                project,
                usage.id,
            )
            num_reservations = sum(
                r.delta for r in reservations if r.delta > 0
            )
            if num_reservations != usage.reserved:
                print(
                    '\t%s: invalid reserved saved=%s actual=%s%s' %
                    (
                        resource_name,
                        usage.reserved,
                        num_reservations,
                        action_msg,
                    )
                )
                discrepancy = True
                if do_fix:
                    usage.reserved = num_reservations

        return discrepancy


class VersionCommands(object):
    """Class for exposing the codebase version."""

    def __init__(self):
        pass

    def list(self):
        print(version.version_string())

    def __call__(self):
        self.list()


class VolumeCommands(object):
    """Methods for dealing with a cloud in an odd state."""

    @args('volume_id',
          help='Volume ID to be deleted')
    def delete(self, volume_id):
        """Delete a volume, bypassing the check that it must be available."""
        ctxt = context.get_admin_context()
        volume = objects.Volume.get_by_id(ctxt, volume_id)
        host = volume_utils.extract_host(volume.host) if volume.host else None

        if not host:
            print(_("Volume not yet assigned to host."))
            print(_("Deleting volume from database and skipping rpc."))
            volume.destroy()
            return

        if volume.status == 'in-use':
            print(_("Volume is in-use."))
            print(_("Detach volume from instance and then try again."))
            return

        rpc.init(CONF)
        rpcapi = volume_rpcapi.VolumeAPI()
        rpcapi.delete_volume(ctxt, volume)

    @args('--currenthost', required=True, help='Existing volume host name in '
                                               'the format host@backend#pool')
    @args('--newhost', required=True, help='New volume host name in the '
                                           'format host@backend#pool')
    def update_host(self, currenthost, newhost):
        """Modify the host name associated with a volume.

        Particularly to recover from cases where one has moved
        their Cinder Volume node, or modified their backend_name in a
        multi-backend config.
        """
        ctxt = context.get_admin_context()
        volumes = db.volume_get_all_by_host(ctxt,
                                            currenthost)
        for v in volumes:
            db.volume_update(ctxt, v['id'],
                             {'host': newhost})


class ConfigCommands(object):
    """Class for exposing the flags defined by flag_file(s)."""

    def __init__(self):
        pass

    @args('param', nargs='?', default=None,
          help='Configuration parameter to display (default: %(default)s)')
    def list(self, param=None):
        """List parameters configured for cinder.

        Lists all parameters configured for cinder unless an optional argument
        is specified.  If the parameter is specified we only print the
        requested parameter.  If the parameter is not found an appropriate
        error is produced by .get*().
        """
        param = param and param.strip()
        if param:
            print('%s = %s' % (param, CONF.get(param)))
        else:
            for key, value in CONF.items():
                print('%s = %s' % (key, value))


class BackupCommands(object):
    """Methods for managing backups."""

    def list(self):
        """List all backups.

        List all backups (including ones in progress) and the host
        on which the backup operation is running.
        """
        ctxt = context.get_admin_context()
        backups = objects.BackupList.get_all(ctxt)

        hdr = "%-32s\t%-32s\t%-32s\t%-24s\t%-24s\t%-12s\t%-12s\t%-12s\t%-12s"
        print(hdr % (_('ID'),
                     _('User ID'),
                     _('Project ID'),
                     _('Host'),
                     _('Name'),
                     _('Container'),
                     _('Status'),
                     _('Size'),
                     _('Object Count')))

        res = "%-32s\t%-32s\t%-32s\t%-24s\t%-24s\t%-12s\t%-12s\t%-12d\t%-12d"
        for backup in backups:
            object_count = 0
            if backup['object_count'] is not None:
                object_count = backup['object_count']
            print(res % (backup['id'],
                         backup['user_id'],
                         backup['project_id'],
                         backup['host'],
                         backup['display_name'],
                         backup['container'],
                         backup['status'],
                         backup['size'],
                         object_count))

    @args('--currenthost', required=True, help='Existing backup host name')
    @args('--newhost', required=True, help='New backup host name')
    def update_backup_host(self, currenthost, newhost):
        """Modify the host name associated with a backup.

        Particularly to recover from cases where one has moved
        their Cinder Backup node, and not set backup_use_same_backend.
        """
        ctxt = context.get_admin_context()
        backups = objects.BackupList.get_all_by_host(ctxt, currenthost)
        for bk in backups:
            bk.host = newhost
            bk.save()


class BaseCommand(object):
    @staticmethod
    def _normalize_time(time_field):
        return time_field and timeutils.normalize_time(time_field)

    @staticmethod
    def _state_repr(is_up):
        return ':-)' if is_up else 'XXX'


class ServiceCommands(BaseCommand):
    """Methods for managing services."""
    def list(self):
        """Show a list of all cinder services."""
        ctxt = context.get_admin_context()
        services = objects.ServiceList.get_all(ctxt)
        print_format = "%-16s %-36s %-16s %-10s %-5s %-20s %-12s %-15s %-36s"
        print(print_format % (_('Binary'),
                              _('Host'),
                              _('Zone'),
                              _('Status'),
                              _('State'),
                              _('Updated At'),
                              _('RPC Version'),
                              _('Object Version'),
                              _('Cluster')))
        for svc in services:
            art = self._state_repr(svc.is_up)
            status = 'disabled' if svc.disabled else 'enabled'
            updated_at = self._normalize_time(svc.updated_at)
            rpc_version = svc.rpc_current_version
            object_version = svc.object_current_version
            cluster = svc.cluster_name or ''
            print(print_format % (svc.binary, svc.host,
                                  svc.availability_zone, status, art,
                                  updated_at, rpc_version, object_version,
                                  cluster))

    @args('binary', type=str,
          help='Service to delete from the host.')
    @args('host_name', type=str,
          help='Host from which to remove the service.')
    def remove(self, binary, host_name):
        """Completely removes a service."""
        ctxt = context.get_admin_context()
        try:
            svc = objects.Service.get_by_args(ctxt, host_name, binary)
            svc.destroy()
        except exception.ServiceNotFound as e:
            print(_("Host not found. Failed to remove %(service)s"
                    " on %(host)s.") %
                  {'service': binary, 'host': host_name})
            print(u"%s" % e.args)
            return 2
        print(_("Service %(service)s on host %(host)s removed.") %
              {'service': binary, 'host': host_name})


class ClusterCommands(BaseCommand):
    """Methods for managing clusters."""
    def list(self):
        """Show a list of all cinder services."""
        ctxt = context.get_admin_context()
        clusters = objects.ClusterList.get_all(ctxt, services_summary=True)
        print_format = "%-36s %-16s %-10s %-5s %-20s %-7s %-12s %-20s"
        print(print_format % (_('Name'),
                              _('Binary'),
                              _('Status'),
                              _('State'),
                              _('Heartbeat'),
                              _('Hosts'),
                              _('Down Hosts'),
                              _('Updated At')))
        for cluster in clusters:
            art = self._state_repr(cluster.is_up)
            status = 'disabled' if cluster.disabled else 'enabled'
            heartbeat = self._normalize_time(cluster.last_heartbeat)
            updated_at = self._normalize_time(cluster.updated_at)
            print(print_format % (cluster.name, cluster.binary, status, art,
                                  heartbeat, cluster.num_hosts,
                                  cluster.num_down_hosts, updated_at))

    @args('--recursive', action='store_true', default=False,
          help='Delete associated hosts.')
    @args('binary', type=str,
          help='Service to delete from the cluster.')
    @args('cluster-name', type=str, help='Cluster to delete.')
    def remove(self, recursive, binary, cluster_name):
        """Completely removes a cluster."""
        ctxt = context.get_admin_context()
        try:
            cluster = objects.Cluster.get_by_id(ctxt, None, name=cluster_name,
                                                binary=binary,
                                                get_services=recursive)
        except exception.ClusterNotFound:
            print(_("Couldn't remove cluster %s because it doesn't exist.") %
                  cluster_name)
            return 2

        if recursive:
            for service in cluster.services:
                service.destroy()

        try:
            cluster.destroy()
        except exception.ClusterHasHosts:
            print(_("Couldn't remove cluster %s because it still has hosts.") %
                  cluster_name)
            return 2

        msg = _('Cluster %s successfully removed.') % cluster_name
        if recursive:
            msg = (_('%(msg)s And %(num)s services from the cluster were also '
                     'removed.') % {'msg': msg, 'num': len(cluster.services)})
        print(msg)

    @args('--full-rename', dest='partial',
          action='store_false', default=True,
          help='Do full cluster rename instead of just replacing provided '
               'current cluster name and preserving backend and/or pool info.')
    @args('current', help='Current cluster name.')
    @args('new', help='New cluster name.')
    def rename(self, partial, current, new):
        """Rename cluster name for Volumes and Consistency Groups.

        Useful when you want to rename a cluster, particularly when the
        backend_name has been modified in a multi-backend config or we have
        moved from a single backend to multi-backend.
        """
        ctxt = context.get_admin_context()

        # Convert empty strings to None
        current = current or None
        new = new or None

        # Update Volumes
        num_vols = objects.VolumeList.include_in_cluster(
            ctxt, new, partial_rename=partial, cluster_name=current)

        # Update Consistency Groups
        num_cgs = objects.ConsistencyGroupList.include_in_cluster(
            ctxt, new, partial_rename=partial, cluster_name=current)

        if num_vols or num_cgs:
            msg = _('Successfully renamed %(num_vols)s volumes and '
                    '%(num_cgs)s consistency groups from cluster %(current)s '
                    'to %(new)s')
            print(msg % {'num_vols': num_vols, 'num_cgs': num_cgs, 'new': new,
                         'current': current})
        else:
            msg = _('No volumes or consistency groups exist in cluster '
                    '%(current)s.')
            print(msg % {'current': current})
            return 2


class ConsistencyGroupCommands(object):
    """Methods for managing consistency groups."""

    @args('--currenthost', required=True, help='Existing CG host name')
    @args('--newhost', required=True, help='New CG host name')
    def update_cg_host(self, currenthost, newhost):
        """Modify the host name associated with a Consistency Group.

        Particularly to recover from cases where one has moved
        a host from single backend to multi-backend, or changed the host
        configuration option, or modified the backend_name in a multi-backend
        config.
        """

        ctxt = context.get_admin_context()
        groups = objects.ConsistencyGroupList.get_all(
            ctxt, {'host': currenthost})
        for gr in groups:
            gr.host = newhost
            gr.save()


class UtilCommands(object):
    """Generic utils."""

    @staticmethod
    def _get_resources_locks():
        """Get all vol/snap/backup file lock paths."""
        backup_locks_prefix = 'cinder-cleanup_incomplete_backups_'
        oslo_dir = os.path.abspath(cfg.CONF.oslo_concurrency.lock_path)
        filenames = glob.glob(os.path.join(oslo_dir, 'cinder-*'))

        backend_url = cfg.CONF.coordination.backend_url
        if backend_url.startswith('file://'):
            tooz_dir = os.path.abspath(backend_url[7:])
            if tooz_dir != oslo_dir:
                filenames += glob.glob(os.path.join(tooz_dir, 'cinder-*'))

        volumes = collections.defaultdict(list)
        snapshots = collections.defaultdict(list)
        backups = collections.defaultdict(list)
        matcher = re.compile('.*?([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-'
                             '[0-9a-f]{4}-[0-9a-f]{12}).*?', re.IGNORECASE)
        for filename in filenames:
            basename = os.path.basename(filename)
            match = matcher.match(basename)
            if match:
                dest = snapshots if 'snapshot' in basename else volumes
                res_id = match.group(1)
                dest[res_id].append(filename)
            elif basename.startswith(backup_locks_prefix):
                pgrp = basename[34:]
                backups[pgrp].append(filename)

        return volumes, snapshots, backups

    def _exclude_running_backups(self, backups):
        """Remove backup entries from the dict for running backup services."""
        for backup_pgrp in list(backups.keys()):
            # The PGRP is the same as the PID of the parent process, so we know
            # the lock could be in use if the process is running and it's the
            # cinder-backup command (the PID could have been reused).
            cmdline_file = os.path.join('/proc', backup_pgrp, 'cmdline')
            try:
                with open(cmdline_file, 'r') as f:
                    if 'cinder-backup' in f.read():
                        del backups[backup_pgrp]
            except FileNotFoundError:
                continue
            except Exception:
                # Unexpected error, leaving the lock file just in case
                del backups[backup_pgrp]

    @args('--services-offline', dest='online',
          action='store_false', default=True,
          help='All locks can be deleted as Cinder services are not running.')
    def clean_locks(self, online):
        """Clean file locks for vols, snaps, and backups on the current host.

        Should be run on any host where we are running a Cinder service (API,
        Scheduler, Volume, Backup) and can be run with the Cinder services
        running or stopped.

        If the services are running it will check existing resources in the
        Cinder database in order to know which resources are still available
        (it's not safe to remove their file locks) and will only remove the
        file locks for the resources that are no longer present.  Deleting
        locks while the services are offline is faster as there's no need to
        check the database.

        For backups, the way to know if we can remove the startup lock is by
        checking if the PGRP in the file name is currently running
        cinder-backup.

        Default assumes that services are online, must pass
        ``--services-offline`` to specify that they are offline.

        Doesn't clean DLM locks (except when using file locks), as those don't
        leave lock leftovers.
        """
        self.ctxt = context.get_admin_context()
        # Find volume and snapshots ids, and backups PGRP based on the existing
        # file locks
        volumes, snapshots, backups = self._get_resources_locks()

        # If services are online we cannot delete locks for existing resources
        if online:
            # We don't want to delete file locks for existing resources
            volumes = {vol_id: files for vol_id, files in volumes.items()
                       if not objects.Volume.exists(self.ctxt, vol_id)}
            snapshots = {snap_id: files for snap_id, files in snapshots.items()
                         if not objects.Snapshot.exists(self.ctxt, snap_id)}
            self._exclude_running_backups(backups)

        def _err(filename: str, exc: Exception) -> None:
            print('Failed to cleanup lock %(name)s: %(exc)s',
                  {'name': filename, 'exc': exc})

        # Now clean
        for filenames in itertools.chain(volumes.values(),
                                         snapshots.values(),
                                         backups.values()):
            for filename in filenames:
                try:
                    os.remove(filename)
                except OSError as exc:
                    if (exc.errno != errno.ENOENT):
                        _err(filename, exc)
                except Exception as exc:
                    _err(filename, exc)


CATEGORIES = {
    'backup': BackupCommands,
    'config': ConfigCommands,
    'cluster': ClusterCommands,
    'cg': ConsistencyGroupCommands,
    'db': DbCommands,
    'host': HostCommands,
    'quota': QuotaCommands,
    'service': ServiceCommands,
    'version': VersionCommands,
    'volume': VolumeCommands,
    'util': UtilCommands,
}


def methods_of(obj):
    """Return non-private methods from an object.

    Get all callable methods of an object that don't start with underscore
    :return: a list of tuples of the form (method_name, method)
    """
    result = []
    for i in dir(obj):
        if isinstance(getattr(obj, i),
                      collections_abc.Callable) and not i.startswith('_'):
            result.append((i, getattr(obj, i)))
    return result


def missing_action(help_func):
    def wrapped():
        help_func()
        exit(2)
    return wrapped


def add_command_parsers(subparsers):
    for category in sorted(CATEGORIES):
        command_object = CATEGORIES[category]()

        parser = subparsers.add_parser(category)
        parser.set_defaults(command_object=command_object)
        parser.set_defaults(action_fn=missing_action(parser.print_help))

        category_subparsers = parser.add_subparsers(dest='action')

        for (action, action_fn) in methods_of(command_object):
            parser = category_subparsers.add_parser(action)

            action_kwargs = []
            for args, kwargs in getattr(action_fn, 'args', []):
                parser.add_argument(*args, **kwargs)

            parser.set_defaults(action_fn=action_fn)
            parser.set_defaults(action_kwargs=action_kwargs)


category_opt = cfg.SubCommandOpt('category',
                                 title='Command categories',
                                 handler=add_command_parsers)


def get_arg_string(args):
    if args[0] == '-':
        # (Note)zhiteng: args starts with FLAGS.oparser.prefix_chars
        # is optional args. Notice that cfg module takes care of
        # actual ArgParser so prefix_chars is always '-'.
        if args[1] == '-':
            # This is long optional arg
            args = args[2:]
        else:
            args = args[1:]  # pylint: disable=E1136

    # We convert dashes to underscores so we can have cleaner optional arg
    # names
    if args:
        args = args.replace('-', '_')

    return args


def fetch_func_args(func):
    fn_kwargs = {}
    for args, kwargs in getattr(func, 'args', []):
        # Argparser `dest` configuration option takes precedence for the name
        arg = kwargs.get('dest') or get_arg_string(args[0])
        fn_kwargs[arg] = getattr(CONF.category, arg)

    return fn_kwargs


def main():
    objects.register_all()
    """Parse options and call the appropriate class/method."""
    CONF.register_cli_opt(category_opt)
    script_name = sys.argv[0]
    if len(sys.argv) < 2:
        print(_("\nOpenStack Cinder version: %(version)s\n") %
              {'version': version.version_string()})
        print(script_name + " category action [<args>]")
        print(_("Available categories:"))
        for category in CATEGORIES:
            print(_("\t%s") % category)
        sys.exit(2)

    try:
        CONF(sys.argv[1:], project='cinder',
             version=version.version_string())
        logging.setup(CONF, "cinder")
        python_logging.captureWarnings(True)
    except cfg.ConfigDirNotFoundError as details:
        print(_("Invalid directory: %s") % details)
        sys.exit(2)
    except cfg.ConfigFilesNotFoundError as e:
        cfg_files = e.config_files
        print(_("Failed to read configuration file(s): %s") % cfg_files)
        sys.exit(2)

    fn = CONF.category.action_fn
    fn_kwargs = fetch_func_args(fn)
    fn(**fn_kwargs)



#    (c) Copyright 2019 Brocade, a Broadcom Company
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Brocade Zone Connector Factory is responsible to dynamically create the
connection object based on the configuration
"""

from oslo_log import log as logging
from oslo_utils import importutils

from cinder.zonemanager.drivers.brocade import fc_zone_constants

LOG = logging.getLogger(__name__)


class BrcdFCZoneFactory(object):

    def __init__(self):
        self.sb_conn_map = {}

    def get_connector(self, fabric, sb_connector):
        """Returns Device Connector.

        Factory method to create and return
        correct SB connector object based on the protocol
        """

        fabric_ip = fabric.safe_get('fc_fabric_address')
        client = self.sb_conn_map.get(fabric_ip)

        if not client:

            fabric_user = fabric.safe_get('fc_fabric_user')
            fabric_pwd = fabric.safe_get('fc_fabric_password')
            fabric_port = fabric.safe_get('fc_fabric_port')
            fc_vfid = fabric.safe_get('fc_virtual_fabric_id')
            fabric_ssh_cert_path = fabric.safe_get('fc_fabric_ssh_cert_path')

            LOG.debug("Client not found. Creating connection client for"
                      " %(ip)s with %(connector)s protocol "
                      "for the user %(user)s at port %(port)s.",
                      {'ip': fabric_ip,
                       'connector': sb_connector,
                       'user': fabric_user,
                       'port': fabric_port,
                       'vf_id': fc_vfid})

            if sb_connector.lower() in (fc_zone_constants.REST_HTTP,
                                        fc_zone_constants.REST_HTTPS):
                client = importutils.import_object(
                    "cinder.zonemanager.drivers.brocade."
                    "brcd_rest_fc_zone_client.BrcdRestFCZoneClient",
                    ipaddress=fabric_ip,
                    username=fabric_user,
                    password=fabric_pwd,
                    port=fabric_port,
                    vfid=fc_vfid,
                    protocol=sb_connector
                )
            elif sb_connector.lower() in (fc_zone_constants.HTTP,
                                          fc_zone_constants.HTTPS):
                client = importutils.import_object(
                    "cinder.zonemanager.drivers.brocade."
                    "brcd_http_fc_zone_client.BrcdHTTPFCZoneClient",
                    ipaddress=fabric_ip,
                    username=fabric_user,
                    password=fabric_pwd,
                    port=fabric_port,
                    vfid=fc_vfid,
                    protocol=sb_connector
                )
            else:
                client = importutils.import_object(
                    "cinder.zonemanager.drivers.brocade."
                    "brcd_fc_zone_client_cli.BrcdFCZoneClientCLI",
                    ipaddress=fabric_ip,
                    username=fabric_user,
                    password=fabric_pwd,
                    key=fabric_ssh_cert_path,
                    port=fabric_port
                )
            self.sb_conn_map.update({fabric_ip: client})
        return client
